"""
Main entry point for the AMS CLI and scripting interfaces.
"""

import cProfile
import io
import logging
import os
import platform
import pstats
import sys
from functools import partial
from subprocess import call
from time import sleep
from typing import Optional, Union
import textwrap

from ._version import get_versions

from andes.main import _find_cases
from andes.shared import Pool, Process, coloredlogs, unittest, NCPUS_PHYSICAL
from andes.utils.misc import elapsed, is_interactive

import ams
from ams.routines import routine_cli
from ams.system import System
from ams.utils.paths import get_config_path, get_log_dir, tests_root
from ams.shared import copyright_msg

logger = logging.getLogger(__name__)


def config_logger(stream_level=logging.INFO, *,
                  stream=True,
                  file=True,
                  log_file='ams.log',
                  log_path=None,
                  file_level=logging.DEBUG,
                  ):
    """
    Configure an AMS logger with a `FileHandler` and a `StreamHandler`.

    This function is called at the beginning of ``ams.main.main()``.
    Updating ``stream_level`` and ``file_level`` is now supported.

    Parameters
    ----------
    stream : bool, optional
        Create a `StreamHandler` for `stdout` if ``True``.
        If ``False``, the handler will not be created.
    file : bool, optionsl
        True if logging to ``log_file``.
    log_file : str, optional
        Logg file name for `FileHandler`, ``'ams.log'`` by default.
        If ``None``, the `FileHandler` will not be created.
    log_path : str, optional
        Path to store the log file. By default, the path is generated by
        get_log_dir() in utils.misc.
    stream_level : {10, 20, 30, 40, 50}, optional
        `StreamHandler` verbosity level.
    file_level : {10, 20, 30, 40, 50}, optional
        `FileHandler` verbosity level.

    Returns
    -------
    None

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    lg = logging.getLogger('ams')
    lg.setLevel(logging.DEBUG)

    if log_path is None:
        log_path = get_log_dir()

    sh_formatter_str = '%(message)s'
    if stream_level == 1:
        sh_formatter_str = '%(name)s:%(lineno)d - %(levelname)s - %(message)s'
        stream_level = 10

    sh_formatter = logging.Formatter(sh_formatter_str)
    if len(lg.handlers) == 0:

        # create a StreamHandler
        if stream is True:
            sh = logging.StreamHandler()
            sh.setFormatter(sh_formatter)
            sh.setLevel(stream_level)
            lg.addHandler(sh)

        # file handler for level DEBUG and up
        if file is True and (log_file is not None):
            log_full_path = os.path.join(log_path, log_file)
            fh_formatter = logging.Formatter('%(process)d: %(asctime)s - %(name)s - %(levelname)s - %(message)s')
            fh = logging.FileHandler(log_full_path)
            fh.setLevel(file_level)
            fh.setFormatter(fh_formatter)
            lg.addHandler(fh)

        globals()['logger'] = lg

    else:
        # update the handlers
        set_logger_level(logger, logging.StreamHandler, stream_level)
        set_logger_level(logger, logging.FileHandler, file_level)

    if not is_interactive():
        coloredlogs.install(logger=lg, level=stream_level, fmt=sh_formatter_str)


def load(case, setup=True,
         use_input_path=True,
         **kwargs):
    """
    Load a case and set up a system without running routine.
    Return a system.

    Takes other kwargs recognizable by ``System``,
    such as ``addfile``, ``input_path``, and ``no_putput``.

    Parameters
    ----------
    case: str
        Path to the test case
    setup : bool, optional
        Call `System.setup` after loading
    use_input_path : bool, optional
        True to use the ``input_path`` argument to behave
        the same as ``ams.main.run``.

    Warnings
    --------
    If one need to add devices in addition to these from the case
    file, do ``setup=False`` and call ``System.add()`` to add devices.
    When done, manually invoke ``setup()`` to set up the system.

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    if use_input_path:
        input_path = kwargs.get('input_path', '')
        case = _find_cases(case, input_path)
        if len(case) > 1:
            logger.error("`ams.load` does not support mulitple cases.")
            return None
        elif len(case) == 0:
            logger.error("No valid case found.")
            return None
        case = case[0]

    system = System(case, **kwargs)

    if not ams.io.parse(system):
        return None

    if setup:
        system.setup()
    return system


def run_case(case, *, routine='pflow', profile=False,
             convert='', convert_all='', add_book=None,
             **kwargs):
    """
    Run single simulation case for the given full path.
    Use ``run`` instead of ``run_case`` whenever possible.

    Argument ``input_path`` will not be prepended to ``case``.

    Arguments recognizable by ``load`` can be passed to ``run_case``.

    Parameters
    ----------
    case : str
        Full path to the test case
    routine : str, ('pflow', 'tds', 'eig')
        Computation routine to run
    profile : bool, optional
        True to enable profiler
    convert : str, optional
        Format name for case file conversion.
    convert_all : str, optional
        Format name for case file conversion, output
        sheets for all available devices.
    add_book : str, optional
        Name of the device to be added to an excel case
        as a new sheet.

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    pr = cProfile.Profile()
    # enable profiler if requested
    if profile is True:
        pr.enable()

    system = load(case,
                  use_input_path=False,
                  **kwargs)

    if system is None:
        return None

    skip_empty = True
    overwrite = None
    # convert to xlsx and process `add-book` option
    if add_book is not None:
        convert = 'xlsx'
        overwrite = True
    if convert_all != '':
        convert = 'xlsx'
        skip_empty = False

    # convert to the requested format
    if convert != '':
        ams.io.dump(system, convert, overwrite=overwrite, skip_empty=skip_empty,
                    add_book=add_book)
        return system

    # run the requested routine
    if routine is not None:
        if isinstance(routine, str):
            routine = [routine]

        if system.is_setup:
            for r in routine:
                system.__dict__[routine_cli[r.lower()]].run(**kwargs)
        else:
            logger.error("System is not set up. Routines cannot continue.")

    # Disable profiler and output results
    if profile:
        pr.disable()

        if system.files.no_output:
            nlines = 40
            s = io.StringIO()
            ps = pstats.Stats(pr, stream=sys.stdout).sort_stats('cumtime')
            ps.print_stats(nlines)
            logger.info(s.getvalue())
            s.close()
        else:
            nlines = 999
            with open(system.files.prof, 'w') as s:
                ps = pstats.Stats(pr, stream=s).sort_stats('cumtime')
                ps.print_stats(nlines)
                ps.dump_stats(system.files.prof_raw)
            logger.info('cProfile text data written to "%s".', system.files.prof)
            logger.info('cProfile raw data written to "%s". View with tool `snakeviz`.', system.files.prof_raw)

    return system


def _run_mp_proc(cases, ncpu=NCPUS_PHYSICAL, **kwargs):
    """
    Run multiprocessing with `Process`.

    Return values from `run_case` are not preserved. Always return `True` when done.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    # start processes
    jobs = []
    for idx, file in enumerate(cases):
        job = Process(name=f'Process {idx:d}', target=run_case, args=(file,), kwargs=kwargs)
        jobs.append(job)
        job.start()
        start_msg = f'Process {idx:d} for "{file:s}" started.'
        print(start_msg)
        logger.debug(start_msg)
        if (idx % ncpu == ncpu - 1) or (idx == len(cases) - 1):
            sleep(0.1)
            for job in jobs:
                job.join()
            jobs = []

    return True


def _run_mp_pool(cases, ncpu=NCPUS_PHYSICAL, verbose=logging.INFO, **kwargs):
    """
    Run multiprocessing jobs using Pool.

    This function returns all System instances in a list, but requires longer computation time.

    Parameters
    ----------
    ncpu : int, optional = os.cpu_cout()
        Number of cpu cores to use in parallel
    mp_verbose : 10 - 50
        Verbosity level during multiprocessing
    verbose : 10, 20, 30, 40, 50
        Verbosity level outside multiprocessing

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    pool = Pool(ncpu)
    print("Cases are processed in the following order:")
    print('\n'.join([f'"{name}"' for name in cases]))

    ret = pool.map(partial(run_case,
                           verbose=verbose,
                           remove_pycapsule=True,
                           autogen_stale=False,
                           **kwargs),
                   cases)

    # FIXME: does following code work in AMS?
    # # fix address for in-place arrays
    # for ss in ret:
    #     fix_view_arrays(ss)

    return ret


def run(filename, input_path='', verbose=20, mp_verbose=30,
        ncpu=NCPUS_PHYSICAL, pool=False,
        cli=False, shell=False, **kwargs):
    """
    Entry point to run AMS routines.

    Parameters
    ----------
    filename : str
        file name (or pattern)
    input_path : str, optional
        input search path
    verbose : int, 10 (DEBUG), 20 (INFO), 30 (WARNING), 40 (ERROR), 50 (CRITICAL)
        Verbosity level. If ``config_logger`` is called prior to ``run``,
        this option will be ignored.
    mp_verbose : int
        Verbosity level for multiprocessing tasks
    ncpu : int, optional
        Number of cpu cores to use in parallel
    pool: bool, optional
        Use Pool for multiprocessing to return a list of created Systems.
    kwargs
        Other supported keyword arguments
    cli : bool, optional
        If is running from command-line. If True, returns exit code instead of System
    shell : bool, optional
        If True, enter IPython shell after routine.

    Returns
    -------
    System or exit_code
        An instance of system (if `cli == False`) or an exit code otherwise..

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    if is_interactive() and len(logger.handlers) == 0:
        config_logger(verbose, file=False)

    # put some args back to `kwargs`
    kwargs['input_path'] = input_path
    kwargs['verbose'] = verbose

    cases = _find_cases(filename, input_path)

    system = None
    ex_code = 0

    if len(filename) > 0 and len(cases) == 0:
        ex_code = 1  # file specified but not found

    t0, _ = elapsed()
    if len(cases) == 1:
        system = run_case(cases[0], **kwargs)
    elif len(cases) > 1:
        # FIXME: after standardize code generation, enable following code
        # # import `pycode` to local namespace to avoid a picking issue
        # import_pycode()

        # suppress logging output during multiprocessing
        logger.info('-> Processing %s jobs on %s CPUs.', len(cases), ncpu)
        set_logger_level(logger, logging.StreamHandler, mp_verbose)
        set_logger_level(logger, logging.FileHandler, logging.DEBUG)

        if pool is True:
            system = _run_mp_pool(cases,
                                  ncpu=ncpu,
                                  mp_verbose=mp_verbose,
                                  **kwargs)
        else:
            system = _run_mp_proc(cases,
                                  ncpu=ncpu,
                                  mp_verbose=mp_verbose,
                                  **kwargs)

        # restore command line output when all jobs are done
        set_logger_level(logger, logging.StreamHandler, verbose)

        log_files = find_log_path(logger)
        if len(log_files) > 0:
            log_paths = '\n'.join(log_files)
            print(f'Log saved to "{log_paths}".')

    t0, s0 = elapsed(t0)

    if len(cases) == 1:
        if system is not None:
            ex_code += system.exit_code
        else:
            ex_code += 1
    elif len(cases) > 1:
        if isinstance(system, list):
            for s in system:
                ex_code += s.exit_code

    if len(cases) == 1:
        if ex_code == 0:
            print(f'-> Single process finished in {s0}.')
        else:
            print(f'-> Single process exit with an error in {s0}.')
    elif len(cases) > 1:
        if ex_code == 0:
            print(f'-> Multiprocessing finished in {s0}.')
        else:
            print(f'-> Multiprocessing exit with an error in {s0}.')

    # IPython interactive shell
    if shell is True:
        try:
            from IPython import embed

            # load plotter before entering IPython
            if system is None:
                logger.warning("IPython: The System object has not been created.")
            elif isinstance(system, System):
                logger.info("IPython: Access System object in variable `system`.")
                system.TDS.load_plotter()
            elif isinstance(system, list):
                logger.warning("IPython: System objects stored in list `system`.\n"
                               "Call `TDS.load_plotter()` on each for plotter.")

            embed()
        except ImportError:
            logger.warning("IPython import error. Installed?")

    if cli is True:
        return ex_code

    return system


def set_logger_level(lg, type_to_set, level):
    """
    Set logging level for the given type of handler.
    """

    for h in lg.handlers:
        if isinstance(h, type_to_set):
            h.setLevel(level)


def find_log_path(lg):
    """
    Find the file paths of the FileHandlers.
    """
    out = []
    for h in lg.handlers:
        if isinstance(h, logging.FileHandler):
            out.append(h.baseFilename)
    return out


def misc(edit_config='', save_config='', show_license=False, clean=True, recursive=False,
         overwrite=None, version=False, **kwargs):
    """
    Miscellaneous commands.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    if edit_conf(edit_config):
        return
    if show_license:
        print_license()
        return
    if save_config != '':
        save_conf(save_config, overwrite=overwrite, **kwargs)
        return
    if clean is True:
        remove_output(recursive)
        return

    if demo is True:
        demo(**kwargs)
        return

    if version is True:
        versioninfo()
        return

    logger.info("info: no option specified. Use 'ams misc -h' for help.")


def doc(attribute=None, list_supported=False, config=False, **kwargs):
    """
    Quick documentation from command-line.

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    system = System()
    if attribute is not None:
        if attribute in system.__dict__ and hasattr(system.__dict__[attribute], 'doc'):
            logger.info(system.__dict__[attribute].doc())
        else:
            logger.error('Model <%s> does not exist.', attribute)

    elif list_supported is True:
        logger.info(system.supported_routines())

    else:
        logger.info('info: no option specified. Use \'ams doc -h\' for help.')


def demo(**kwargs):
    """
    TODO: show some demonstrations from CLI.
    """
    raise NotImplementedError("Demos have not been implemented")


def versioninfo():
    """
    Print version info for AMS and dependencies.

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    import numpy as np
    import cvxpy
    import andes
    from ams.shared import installed_solvers

    versions = {'Python': platform.python_version(),
                'ams': get_versions()['version'],
                'andes': andes.__version__,
                'numpy': np.__version__,
                'cvxpy': cvxpy.__version__,
                'solvers': ', '.join(installed_solvers),
                }
    maxwidth = max([len(k) for k in versions.keys()])

    for key, val in versions.items():
        if key == 'solvers':
            val = textwrap.fill(val, width=30)  # Wrap solvers to a new line if too long
            val = val.replace('\n', '\n' + ' ' * (maxwidth + 2))
        print(f"{key: <{maxwidth}}  {val}")


def print_license():
    """
    Print out AMS license to stdout.
    """

    print(f"""
    AMS version {ams.__version__}

    {copyright_msg}

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is included below.
    For further information, see <http://www.gnu.org/licenses/>.
    """)
    return True


def edit_conf(edit_config: Optional[Union[str, bool]] = ''):
    """
    Edit the AMS config file which occurs first in the search path.

    Parameters
    ----------
    edit_config : bool
        If ``True``, try to open up an editor and edit the config file. Otherwise returns.

    Returns
    -------
    bool
        ``True`` is a config file is found and an editor is opened. ``False`` if ``edit_config`` is False.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    ret = False

    # no `edit-config` supplied
    if edit_config == '':
        return ret

    conf_path = get_config_path()

    if conf_path is None:
        logger.info('Config file does not exist. Automatically saving.')
        system = System()
        conf_path = system.save_config()

    logger.info('Editing config file "%s"', conf_path)

    editor = ''
    if edit_config is not None:
        # use `edit_config` as default editor
        editor = edit_config
    else:
        # use the following default editors
        if platform.system() == 'Linux':
            editor = os.environ.get('EDITOR', 'vim')
        elif platform.system() == 'Darwin':
            editor = os.environ.get('EDITOR', 'vim')
        elif platform.system() == 'Windows':
            editor = 'notepad.exe'

    editor_cmd = editor.split()
    editor_cmd.append(conf_path)
    call(editor_cmd)
    ret = True
    return ret


def save_conf(config_path=None, overwrite=None, **kwargs):
    """
    Save the AMS config to a file at the path specified by ``save_config``.
    The save action will not run if ``save_config = ''``.

    Parameters
    ----------
    config_path : None or str, optional, ('' by default)

        Path to the file to save the config file. If the path is an emtpy
        string, the save action will not run. Save to
        `~/.ams/ams.conf` if ``None``.

    Returns
    -------
    bool
        ``True`` is the save action is run. ``False`` otherwise.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    ret = False

    # no ``--save-config ``
    if config_path == '':
        return ret

    if config_path is not None and os.path.isdir(config_path):
        config_path = os.path.join(config_path, 'ams.rc')

    ps = System(**kwargs)
    ps.save_config(config_path, overwrite=overwrite)
    ret = True

    return ret


# TODO: change to AMS output types
def remove_output(recursive=False):
    """
    Remove the outputs generated by AMS, including power flow reports
    ``_out.txt``, time-domain list ``_out.lst`` and data ``_out.dat``,
    eigenvalue analysis report ``_eig.txt``.

    Parameters
    ----------
    recursive : bool
        Recursively clean all subfolders

    Returns
    -------
    bool
        ``True`` is the function body executes with success. ``False``
        otherwise.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """
    found = False
    cwd = os.getcwd()

    if recursive:
        dirs = [x[0] for x in os.walk(cwd)]
    else:
        dirs = (cwd,)

    for d in dirs:
        for file in os.listdir(d):
            if file.endswith('_eig.txt') or \
                    file.endswith('_out.txt'):
                found = True
                try:
                    os.remove(os.path.join(d, file))
                    logger.info('"%s" removed.', os.path.join(d, file))
                except IOError:
                    logger.error('Error removing file "%s".',
                                 os.path.join(d, file))
    if not found:
        logger.info('No output file found in the working directory.')

    return True


def selftest(quick=False, extra=False, **kwargs):
    """
    Run unit tests.

    Notes
    -----
    Copied from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    # map verbosity level from logging to unittest
    vmap = {1: 3, 10: 3, 20: 2, 30: 1, 40: 1, 50: 1}
    verbose = vmap[kwargs.get('verbose', 20)]

    # skip if quick
    quick_skips = ('test_1_docs', 'test_codegen_inc')

    # extra test naming convention
    extra_test = 'extra_test'

    try:
        logger.handlers[0].setLevel(logging.WARNING)
        sys.stdout = open(os.devnull, 'w')  # suppress print statements
    except IndexError:  # logger not set up
        pass

    # discover test cases
    test_directory = tests_root()
    suite = unittest.TestLoader().discover(test_directory)

    # remove codegen for quick mode
    for test_group in suite._tests:
        for test_class in test_group._tests:
            tests_keep = list()

            if not hasattr(test_class, '_tests'):
                continue
            for t in test_class._tests:
                # skip the extra tests if `extra` is not True
                if (extra is not True) and (extra_test in t._testMethodName):
                    continue

                # skip the ones for `quick`
                if quick is True and (t._testMethodName in quick_skips):
                    continue

                tests_keep.append(t)

            test_class._tests = tests_keep

    unittest.TextTestRunner(verbosity=verbose).run(suite)
    sys.stdout = sys.__stdout__



"""
Continuous power flow routine.
"""
import logging

from ams.pypower import runcpf

from ams.io.pypower import system2ppc
from ams.pypower.core import ppoption

from ams.routines.pflow import PFlow

logger = logging.getLogger(__name__)


class CPF(PFlow):
    """
    Continuous power flow.

    Still under development, not ready for use.
    """

    def __init__(self, system, config):
        PFlow.__init__(self, system, config)
        self.info = 'AC continuous power flow'
        self.type = 'PF'

    def solve(self, method=None, **kwargs):
        """
        Solve the CPF using PYPOWER.
        """
        ppc = system2ppc(self.system)
        ppopt = ppoption()
        res, success, sstats = runcpf(casedata=ppc, ppopt=ppopt, **kwargs)
        return res, success, sstats

    # FIXME: unpack results?

    def run(self, force_init=False, no_code=True,
            method='newton', **kwargs):
        """
        Run continuous power flow using PYPOWER.

        Examples
        --------
        >>> ss = ams.load(ams.get_case('matpower/case14.m'))
        >>> ss.CPF.run()

        Parameters
        ----------
        force_init : bool
            Force initialization.
        no_code : bool
            Disable showing code.
        method : str
            Method for solving the power flow.

        Returns
        -------
        exit_code : int
            Exit code of the routine.
        """
        super().run(force_init=force_init,
                    no_code=no_code, method=method,
                    **kwargs, )



"""
Column indices.
"""

from collections import OrderedDict  # NOQA


gen_src = OrderedDict([
    ('GEN_BUS', 0),     # bus number
    ('PG', 1),          # Pg, real power output (MW)
    ('QG', 2),          # Qg, reactive power output (MVAr)
    ('QMAX', 3),        # Qmax, maximum reactive power output at Pmin (MVAr)
    ('QMIN', 4),        # Qmin, minimum reactive power output at Pmin (MVAr)
    ('VG', 5),          # Vg, voltage magnitude setpoint (p.u.)
    ('MBASE', 6),       # mBase, total MVA base of this machine, defaults to baseMVA
    ('GEN_STATUS', 7),  # status, 1 - machine in service, 0 - machine out of service
    ('PMAX', 8),        # Pmax, maximum real power output (MW)
    ('PMIN', 9),        # Pmin, minimum real power output (MW)
    ('PC1', 10),        # Pc1, lower real power output of PQ capability curve (MW)
    ('PC2', 11),        # Pc2, upper real power output of PQ capability curve (MW)
    ('QC1MIN', 12),     # Qc1min, minimum reactive power output at Pc1 (MVAr)
    ('QC1MAX', 13),     # Qc1max, maximum reactive power output at Pc1 (MVAr)
    ('QC2MIN', 14),     # Qc2min, minimum reactive power output at Pc2 (MVAr)
    ('QC2MAX', 15),     # Qc2max, maximum reactive power output at Pc2 (MVAr)
    ('RAMP_AGC', 16),   # ramp rate for load following/AGC (MW/min)
    ('RAMP_10', 17),    # ramp rate for 10 minute reserves (MW)
    ('RAMP_30', 18),    # ramp rate for 30 minute reserves (MW)
    ('RAMP_Q', 19),     # ramp rate for reactive power (2 sec timescale) (MVAr/min)
    ('APF', 20),        # area participation factor
    # -----  OPF Data  -----
    # NOTE: included in opf solution, not necessarily in input
    # NOTE: assume objective function has units, u
    ('MU_PMAX', 21),       # Kuhn-Tucker multiplier on upper Pg limit (u/MW)
    ('MU_PMIN', 22),       # Kuhn-Tucker multiplier on lower Pg limit (u/MW)
    ('MU_QMAX', 23),       # Kuhn-Tucker multiplier on upper Qg limit (u/MVAr)
    ('MU_QMIN', 24),       # Kuhn-Tucker multiplier on lower Qg limit (u/MVAr)
])

dcline_src = OrderedDict([
    ('F_BUS', 0),     # f, "from" bus number
    ('T_BUS', 1),     # t,  "to"  bus number
    ('BR_STATUS', 2),  # initial branch status, 1 - in service, 0 - out of service
    ('PF', 3),        # MW flow at "from" bus ("from" -> "to")
    ('PT', 4),        # MW flow at  "to"  bus ("from" -> "to")
    ('QF', 5),        # MVAr injection at "from" bus ("from" -> "to")
    ('QT', 6),        # MVAr injection at  "to"  bus ("from" -> "to")
    ('VF', 7),        # voltage setpoint at "from" bus (p.u.)
    ('VT', 8),        # voltage setpoint at  "to"  bus (p.u.)
    ('PMIN', 9),      # lower limit on PF (MW flow at "from" end)
    ('PMAX', 10),     # upper limit on PF (MW flow at "from" end)
    ('QMINF', 11),    # lower limit on MVAr injection at "from" bus
    ('QMAXF', 12),    # upper limit on MVAr injection at "from" bus
    ('QMINT', 13),    # lower limit on MVAr injection at  "to"  bus
    ('QMAXT', 14),    # upper limit on MVAr injection at  "to"  bus
    ('LOSS0', 15),    # constant term of linear loss function (MW)
    ('LOSS1', 16),    # linear term of linear loss function (MW)
    ('MU_PMIN', 17),  # Kuhn-Tucker multiplier on lower flow limit at "from" bus (u/MW)
    ('MU_PMAX', 18),  # Kuhn-Tucker multiplier on upper flow limit at "from" bus (u/MW)
    ('MU_QMINF', 19),  # Kuhn-Tucker multiplier on lower Q limit at "from" bus (u/MVAr)
    ('MU_QMAXF', 20),  # Kuhn-Tucker multiplier on upper Q limit at "from" bus (u/MVAr)
    ('MU_QMINT', 21),  # Kuhn-Tucker multiplier on lower Q limit at  "to"  bus (u/MVAr)
    ('MU_QMAXT', 22),  # Kuhn-Tucker multiplier on upper Q limit at  "to"  bus (u/MVAr)
])

cost_src = OrderedDict([
    ('PW_LINEAR', 1),   # cost model, 1 - piecewise linear, 2 - polynomial
    ('POLYNOMIAL', 2),  # cost model, 1 - piecewise linear, 2 - polynomial
    ('MODEL', 0),       # cost model, 1 - piecewise linear, 2 - polynomial
    ('STARTUP', 1),     # startup cost in US dollars
    ('SHUTDOWN', 2),    # shutdown cost in US dollars
    ('NCOST', 3),       # number of cost coefficients to follow
    ('COST', 4),        # cost coefficients, see below
])

bus_src = OrderedDict([
    # --- bus types ---
    ('PQ', 1),          # PQ bus
    ('PV', 2),          # PV bus
    ('REF', 3),         # reference bus
    ('NONE', 4),    # isolated bus
    # --- indices ---
    ('BUS_I', 0),       # bus number (1 to 29997)
    ('BUS_TYPE', 1),    # bus type
    ('PD', 2),          # Pd, real power demand (MW)
    ('QD', 3),          # Qd, reactive power demand (MVAr)
    ('GS', 4),          # Gs, shunt conductance (MW at V = 1.0 p.u.)
    ('BS', 5),          # Bs, shunt susceptance (MVAr at V = 1.0 p.u.)
    ('BUS_AREA', 6),    # area number, 1-100
    ('VM', 7),          # Vm, voltage magnitude (p.u.)
    ('VA', 8),          # Va, voltage angle (degrees)
    ('BASE_KV', 9),     # baseKV, base voltage (kV)
    ('ZONE', 10),       # zone, loss zone (1-999)
    ('VMAX', 11),       # maxVm, maximum voltage magnitude (p.u.)
    ('VMIN', 12),       # minVm, minimum voltage magnitude (p.u.)
    # NOTE: included in opf solution, not necessarily in input
    # NOTE: assume objective function has units, u
    ('LAM_P', 13),      # Lagrange multiplier on real power mismatch (u/MW)
    ('LAM_Q', 14),      # Lagrange multiplier on reactive power mismatch (u/MVAr)
    ('MU_VMAX', 15),    # Kuhn-Tucker multiplier on upper voltage limit (u/p.u.)
    ('MU_VMIN', 16),    # Kuhn-Tucker multiplier on lower voltage limit (u/p.u.)
])

branch_src = OrderedDict([
    ('F_BUS', 0),       # f, from bus number
    ('T_BUS', 1),       # t, to bus number
    ('BR_R', 2),        # r, resistance (p.u.)
    ('BR_X', 3),        # x, reactance (p.u.)
    ('BR_B', 4),        # b, total line charging susceptance (p.u.)
    ('RATE_A', 5),      # rateA, MVA rating A (long term rating)
    ('RATE_B', 6),      # rateB, MVA rating B (short term rating)
    ('RATE_C', 7),      # rateC, MVA rating C (emergency rating)
    ('TAP', 8),         # ratio, transformer off nominal turns ratio
    ('SHIFT', 9),       # angle, transformer phase shift angle (degrees)
    ('BR_STATUS', 10),  # initial branch status, 1 - in service, 0 - out of service
    ('ANGMIN', 11),     # minimum angle difference, angle(Vf) - angle(Vt) (degrees)
    ('ANGMAX', 12),     # maximum angle difference, angle(Vf) - angle(Vt) (degrees)
    # NOTE: included in opf solution, not necessarily in input
    ('PF', 13),         # real power injected at "from" bus (MW)
    ('QF', 14),         # reactive power injected at "from" bus (MVAr)
    ('PT', 15),         # real power injected at "to" bus (MW)
    ('QT', 16),         # reactive power injected at "to" bus (MVAr)
    # NOTE: included in opf solution, not necessarily in input
    # NOTE: assume objective function has units, u
    ('MU_SF', 17),      # Kuhn-Tucker multiplier on MVA limit at "from" bus (u/MVA)
    ('MU_ST', 18),      # Kuhn-Tucker multiplier on MVA limit at "to" bus (u/MVA)
    ('MU_ANGMIN', 19),  # Kuhn-Tucker multiplier on lower angle difference limit (u/degree)
    ('MU_ANGMAX', 20),  # Kuhn-Tucker multiplier on upper angle difference limit (u/degree)
])

area_src = OrderedDict([
    ('AREA_I', 0),    # area number
    ('PRICE_REF_BUS', 1),    # price reference bus for this area
])


class Consts:
    """
    Base class for constants collection.
    """

    def __init__(self, source):
        """
        Parameters
        ----------
        source : OrderedDict
            Dictionary of constants.
        """
        self.init(source)

    def init(self, source):
        for key, value in source.items():
            setattr(self, key, value)


class IDXClass:
    """
    Column indices for named columns.
    """

    def __init__(self):
        """
        Parameters
        ----------
        source : OrderedDict
            Dictionary of column indices.
        """
        self.gen = Consts(gen_src)
        self.dcline = Consts(dcline_src)
        self.cost = Consts(cost_src)
        self.bus = Consts(bus_src)
        self.branch = Consts(branch_src)
        self.area = Consts(area_src)

IDX = IDXClass()



"""
EV model.
"""

from andes.core.param import NumParam, IdxParam
from andes.core.model import ModelData

from ams.core.model import Model


class EV1(ModelData, Model):
    """
    Aggregated EV model for scheduling at transmission level.

    For co-simulation with ADNES, it is expected to be used in
    conjunction with the dynamic models `EV1` or `EV2`.

    Reference:

    [1] J. Wang et al., "Electric Vehicles Charging Time Constrained Deliverable Provision of Secondary
    Frequency Regulation," in IEEE Transactions on Smart Grid, doi: 10.1109/TSG.2024.3356948.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'DG'

        self.bus = IdxParam(model='Bus',
                            info="interface bus idx",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Sn = NumParam(default=100.0, tex_name='S_n',
                           info='device MVA rating',
                           unit='MVA',
                           )
        self.gammap = NumParam(default=1.0,
                               info="P ratio of linked static gen",
                               tex_name=r'\gamma_P'
                               )
        self.gammaq = NumParam(default=1.0,
                               info="Q ratio of linked static gen",
                               tex_name=r'\gamma_Q'
                               )
        self.N = NumParam(default=10000,
                          info="number of related EVs",
                          tex_name='N'
                          )


class EV2(EV1):
    """
    EV aggregation model at transmission level, identical to :ref:`EV1`.
    """

    def __init__(self, system=None, config=None) -> None:
        EV1.__init__(self, system, config)



"""
Service.
"""

import logging
from typing import Callable, Type

import numpy as np
import scipy.sparse as spr

from andes.core.service import BaseService

from ams.opt import Param


logger = logging.getLogger(__name__)


class RBaseService(BaseService, Param):
    """
    Base class for services that are used in a routine.
    Revised from module `andes.core.service.BaseService`.

    Parameters
    ----------
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 no_parse: bool = False,
                 sparse: bool = False,
                 ):
        Param.__init__(self, name=name, unit=unit, info=info,
                       no_parse=no_parse)
        BaseService.__init__(self, name=name, tex_name=tex_name, unit=unit,
                             info=info, vtype=vtype)
        self.export = False
        self.is_group = False
        self.rtn = None
        self.sparse = sparse

    @property
    def shape(self):
        """
        Return the shape of the service.
        """
        if isinstance(self.v, (np.ndarray, spr.csr_matrix)):
            return self.v.shape
        else:
            raise TypeError(f'{self.class_name}: {self.name} is not an array.')

    @property
    def v(self):
        """
        Value of the service.
        """
        return None

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def __repr__(self):
        if self.name is None:
            return f'{self.class_name}: {self.rtn.class_name}'
        else:
            return f'{self.class_name}: {self.rtn.class_name}.{self.name}'


class ValueService(RBaseService):
    """
    Service to store given numeric values.

    Parameters
    ----------
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 name: str,
                 value: np.ndarray,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 no_parse: bool = False,
                 sparse: bool = False,
                 ):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         no_parse=no_parse, sparse=sparse)
        self._v = value

    @property
    def v(self):
        """
        Value of the service.
        """
        if self.sparse:
            return spr.csr_matrix(self._v)
        return self._v


class ROperationService(RBaseService):
    """
    Base calss for operational services used in routine.

    Parameters
    ----------
    u : Callable
        Input.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         no_parse=no_parse, sparse=sparse)
        self.u = u


class LoadScale(ROperationService):
    """
    Get zonal load by scale nodal load given the corresponding load scale factor.

    Parameters
    ----------
    u : Callable
        nodal load.
    sd : Callable
        zonal load factor.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 sd: Callable,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 no_parse: bool = False,
                 sparse: bool = False,
                 ):
        tex_name = tex_name if tex_name is not None else u.tex_name
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, u=u, no_parse=no_parse,
                         sparse=sparse)
        self.sd = sd

    @property
    def v(self):
        sys = self.rtn.system
        u_idx = self.u.get_all_idxes()
        ue = self.u.owner.get(src='u', attr='v', idx=u_idx)
        u_bus = self.u.owner.get(src='bus', attr='v', idx=u_idx)
        u_zone = sys.Bus.get(src='zone', attr='v', idx=u_bus)
        u_yloc = np.array(sys.Zone.idx2uid(u_zone))
        p0s = np.multiply(self.sd.v[:, u_yloc].transpose(),
                          (ue * self.u.v)[:, np.newaxis])
        if self.sparse:
            return spr.csr_matrix(p0s)
        return p0s


class NumOp(ROperationService):
    """
    Perform an operation on a numerical array using the
    function ``fun(u.v, **args)``.

    Note that the scalar output is converted to a 1D array.

    The `rargs` are passed to the input function.

    Parameters
    ----------
    u : Callable
        Input.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    model : str, optional
        Model name.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    expand_dims : int, optional
        Expand the dimensions of the output array along a specified axis.
    array_out : bool, optional
        Whether to force the output to be an array.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 fun: Callable,
                 args: dict = None,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 expand_dims: int = None,
                 array_out=True,
                 no_parse: bool = False,
                 sparse: bool = False,):
        tex_name = tex_name if tex_name is not None else u.tex_name
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype, u=u,
                         no_parse=no_parse, sparse=sparse)
        self.fun = fun
        self.args = {} if args is None else args
        self.rfun = rfun
        self.rargs = {} if rargs is None else rargs
        self.expand_dims = expand_dims
        self.array_out = array_out

    @property
    def v0(self):
        out = self.fun(self.u.v, **self.args)
        if self.array_out:
            if not isinstance(out, np.ndarray):
                out = np.array([out])
        return out

    @property
    def v1(self):
        if self.rfun is not None:
            return self.rfun(self.v0, **self.rargs)
        else:
            return self.v0

    @property
    def v(self):
        if self.expand_dims is not None:
            out = np.expand_dims(self.v1, axis=int(self.expand_dims))
        else:
            out = self.v1
        if self.sparse:
            return spr.csr_matrix(out)
        return out


class NumExpandDim(NumOp):
    """
    Expand the dimensions of the input array along a specified axis
    using NumPy's ``np.expand_dims(u.v, axis=axis)``.

    Parameters
    ----------
    u : Callable
        Input.
    axis : int
        Axis along which to expand the dimensions (default is 0).
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    array_out : bool, optional
        Whether to force the output to be an array.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 axis: int = 0,
                 args: dict = None,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 array_out: bool = True,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=np.expand_dims, args=args,
                         array_out=array_out,
                         no_parse=no_parse, sparse=sparse)
        self.axis = axis

    @property
    def v(self):
        out = self.fun(self.u.v, axis=self.axis, **self.args)
        if self.sparse:
            return spr.csr_matrix(out)
        return out


class NumOpDual(NumOp):
    """
    Performan an operation on two numerical arrays using the
    function ``fun(u.v, u2.v, **args)``.

    Note that the scalar output is converted to a 1D array.

    The optional kwargs are passed to the input function.

    Parameters
    ----------
    u : Callable
        Input.
    u2 : Callable
        Input2.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    expand_dims : int, optional
        Expand the dimensions of the output array along a specified axis.
    array_out : bool, optional
        Whether to force the output to be an array.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 u2: Callable,
                 fun: Callable,
                 args: dict = None,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 expand_dims: int = None,
                 array_out=True,
                 no_parse: bool = False,
                 sparse: bool = False,):
        tex_name = tex_name if tex_name is not None else u.tex_name
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=fun, args=args,
                         rfun=rfun, rargs=rargs,
                         expand_dims=expand_dims,
                         array_out=array_out,
                         no_parse=no_parse, sparse=sparse)
        self.u2 = u2

    @property
    def v0(self):
        out = self.fun(self.u.v, self.u2.v, **self.args)
        if self.array_out:
            if not isinstance(out, np.ndarray):
                out = np.array([out])
        if self.sparse:
            return spr.csr_matrix(out)
        return out


class MinDur(NumOpDual):
    """
    Defined to form minimum on matrix for minimum online/offline
    time constraints used in UC.

    Parameters
    ----------
    u : Callable
        Input, should be a ``Var`` with horizon.
    u2 : Callable
        Input2, should be a ``RParam``.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 u2: Callable,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        tex_name = tex_name if tex_name is not None else u.tex_name
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, u2=u2, fun=None, args=None,
                         rfun=None, rargs=None,
                         expand_dims=None,
                         no_parse=no_parse, sparse=sparse)
        if self.u.horizon is None:
            msg = f'{self.class_name} <{self.name}>.u: <{self.u.name}> '
            msg += 'has no horizon, likely a modeling error.'
            logger.error(msg)

    @property
    def v(self):
        n_gen = self.u.n
        n_ts = self.u.horizon.n
        tout = np.zeros((n_gen, n_ts))
        t = self.rtn.config.t  # scheduling interval

        # minimum online/offline duration
        td = np.ceil(self.u2.v/t).astype(int)

        # Create index arrays for generators and time periods
        i, t = np.meshgrid(np.arange(n_gen), np.arange(n_ts), indexing='ij')
        # Create a mask for valid time periods based on minimum duration
        valid_mask = (t + td[i] <= n_ts)
        tout[i[valid_mask], t[valid_mask]] = 1
        if self.sparse:
            return spr.csr_matrix(tout)
        return tout


class NumHstack(NumOp):
    """
    Repeat an array along the second axis nc times or the length of
    reference array,
    using NumPy's hstack function, where nc is the column number of the
    reference array,
    ``np.hstack([u.v[:, np.newaxis] * ref.shape[1]], **kwargs)``.

    Parameters
    ----------
    u : Callable
        Input array.
    ref : Callable
        Reference array used to determine the number of repetitions.
    name : str, optional
        Instance name.
    tex_name : str, optional
        TeX name.
    unit : str, optional
        Unit.
    info : str, optional
        Description.
    vtype : Type, optional
        Variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 ref: Callable,
                 args: dict = None,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=np.hstack, args=args,
                         rfun=rfun, rargs=rargs,
                         no_parse=no_parse, sparse=sparse)
        self.ref = ref

    @property
    def v0(self):
        nc = 1
        if isinstance(self.ref.v, (list, tuple)):
            nc = len(self.ref.v)
        elif hasattr(self.ref, "shape"):
            nc = self.ref.shape[1]
        else:
            raise AttributeError(f"{self.rtn.class_name}: ref {self.ref.name} has no attribute shape nor length.")
        return self.fun([self.u.v[:, np.newaxis]] * nc,
                        **self.args)


class ZonalSum(NumOp):
    """
    Build zonal sum matrix for a vector in the shape of collection model,
    ``Area`` or ``Zone``.
    The value array is in the shape of (nr, nc), where nr is the length of
    rid instance idx, and nc is the length of the cid value.

    In an IEEE-14 Bus system, we have the zonal definition by the
    ``Zone`` model. Suppose in it we have two regions, "ZONE1" and
    "ZONE2".

    Follwing it, we have a zonal SFR requirement model ``SFR`` that
    defines the zonal reserve requirements for each zone.

    All 14 buses are classified to a zone by the `IdxParam` ``zone``,
    and the 5 generators are connected to buses
    (idx): [2, 3, 1, 6, 8], and the zone of these generators are thereby:
    ['ZONE1', 'ZONE1', 'ZONE2', 'ZONE2', 'ZONE1'].

    In the `RTED` model, we have the Vars ``pru`` and ``prd`` in the
    shape of generators.

    Then, the Zone model has idx ['ZONE1', 'ZONE2'], and the ``gsm`` value
    will be [[1, 1, 0, 0, 1], [0, 0, 1, 1, 0]].

    Finally, the zonal reserve requirements can be formulated as
    constraints in the optimization problem: "gsm @ pru <= du" and
    "gsm @ prd <= dd".

    See ``gsm`` definition in :py:mod:`ams.routines.rted.RTEDModel` for
    more details.

    Parameters
    ----------
    u : Callable
        Input.
    zone : str
        Zonal model name, e.g., "Area" or "Zone".
    name : str
        Instance name.
    tex_name : str
        TeX name.
    unit : str
        Unit.
    info : str
        Description.
    vtype : Type
        Variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 zone: str,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=None, args={},
                         rfun=rfun, rargs=rargs,
                         no_parse=no_parse, sparse=sparse)
        self.zone = zone

    @property
    def v0(self):
        try:
            zone_mdl = getattr(self.rtn.system, self.zone)
        except AttributeError:
            raise AttributeError(f'Zonal model <{self.zone}> not found.')
        ridx = None
        try:
            ridx = zone_mdl.idx.v
        except AttributeError:
            ridx = zone_mdl.get_all_idxes()

        row, col = np.meshgrid(self.u.v, ridx)
        # consistency check
        is_subset = set(self.u.v).issubset(set(ridx))
        if not is_subset:
            raise ValueError(f'{self.u.model} contains undefined zone, likey a data error.')
        result = (row == col).astype(int)

        return result


class VarSelect(NumOp):
    """
    A numerical matrix to select a subset of a 2D variable,
    ``u.v[:, idx]``.

    For example, if need to select Energy Storage output
    power from StaticGen `pg`, following definition can be used:
    ```python
    class RTED:
    ...
    self.ce = VarSelect(u=self.pg, indexer='genE')
    ...
    ```

    Parameters
    ----------
    u : Callable
        The input matrix variable.
    indexer: str
        The name of the indexer source.
    gamma : str, optional
        The name of the indexer gamma.
    name : str, optional
        The name of the instance.
    tex_name : str, optional
        The TeX name for the instance.
    unit : str, optional
        The unit of the output.
    info : str, optional
        A description of the operation.
    vtype : Type, optional
        The variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    array_out : bool, optional
        Whether to force the output to be an array.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 indexer: str,
                 gamma: str = None,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 array_out: bool = True,
                 no_parse: bool = False,
                 sparse: bool = False,
                 **kwargs):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype, u=u, fun=None,
                         rfun=rfun, rargs=rargs, array_out=array_out,
                         no_parse=no_parse, sparse=sparse,
                         **kwargs)
        self.indexer = indexer
        self.gamma = gamma

    @property
    def v0(self):
        # FIXME: what if reference source has no idx?
        # data consistency check
        indexer = getattr(self.rtn, self.indexer)
        err_msg = f'Indexer source {indexer.model} has no {indexer.src}.'
        group = model = None
        if indexer.model in self.rtn.system.groups.keys():
            group = self.rtn.system.groups[indexer.model]
            group_idx = group.get_all_idxes()
            try:
                ref = group.get(src=indexer.src, attr='v', idx=group_idx)
            except AttributeError:
                raise AttributeError(err_msg)
        elif indexer.model in self.rtn.system.models.keys():
            model = self.rtn.system.models[indexer.model]
            try:
                ref = model.get(src=indexer.src, attr='v', idx=model.idx.v)
            except AttributeError:
                raise AttributeError(err_msg)
        else:
            raise AttributeError(f'Indexer source model {indexer.model} has no ref.')

        try:
            uidx = self.u.get_all_idxes()
        except AttributeError:
            raise AttributeError(f'Input {self.u.name} has no idx, likey a modeling error.')

        is_empty = len(ref) == 0
        if is_empty:
            raise ValueError(f'{indexer.model} contains no input, likey a data error.')

        is_subset = set(ref).issubset(set(uidx))
        if not is_subset:
            raise ValueError(f'{indexer.model} contains undefined {indexer.src}, likey a data error.')

        row, col = np.meshgrid(uidx, ref)
        out = (row == col).astype(int)
        if self.gamma:
            vgamma = getattr(self.rtn, self.gamma)
            out = vgamma.v[:, np.newaxis] * out
        return out


class VarReduction(NumOp):
    """
    A numerical matrix to reduce a 2D variable to 1D,
    ``np.fun(shape=(1, u.n))``.

    Parameters
    ----------
    u : Callable
        The input matrix variable.
    fun : Callable
        The reduction function that takes a shape argument (1D shape) as input.
    name : str, optional
        The name of the instance.
    tex_name : str, optional
        The TeX name for the instance.
    unit : str, optional
        The unit of the output.
    info : str, optional
        A description of the operation.
    vtype : Type, optional
        The variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 fun: Callable,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=None, rfun=rfun, rargs=rargs,
                         no_parse=no_parse, sparse=sparse,)
        self.fun = fun

    @property
    def v0(self):
        return self.fun(shape=(1, self.u.n))


class RampSub(NumOp):
    """
    Build a substraction matrix for a 2D variable in the shape (nr, nr-1),
    where nr is the rows of the input.

    This can be used for generator ramping constraints in multi-period
    optimization problems.

    The subtraction matrix is constructed as follows:
    ``np.eye(nr, nc, k=-1) - np.eye(nr, nc, k=0)``.

    Parameters
    ----------
    u : Callable
        Input.
    horizon : Callable
        Horizon reference.
    name : str
        Instance name.
    tex_name : str
        TeX name.
    unit : str
        Unit.
    info : str
        Description.
    vtype : Type
        Variable type.
    rfun : Callable, optional
        Function to apply to the output of ``fun``.
    rargs : dict, optional
        Keyword arguments to pass to ``rfun``.
    no_parse: bool, optional
        True to skip parsing the service.
    sparse: bool, optional
        True to return output as scipy csr_matrix.
    """

    def __init__(self,
                 u: Callable,
                 name: str = None,
                 tex_name: str = None,
                 unit: str = None,
                 info: str = None,
                 vtype: Type = None,
                 rfun: Callable = None,
                 rargs: dict = None,
                 no_parse: bool = False,
                 sparse: bool = False,):
        super().__init__(name=name, tex_name=tex_name, unit=unit,
                         info=info, vtype=vtype,
                         u=u, fun=None, rfun=rfun, rargs=rargs,
                         no_parse=no_parse, sparse=sparse,)

    @property
    def v0(self):
        return self.v

    @property
    def v1(self):
        return self.v

    @property
    def v(self):
        nr = self.u.horizon.n
        out = np.eye(nr, nr-1, k=-1) - np.eye(nr, nr-1, k=0)
        if self.sparse:
            return spr.csr_matrix(out)
        return out



"""
Test interface.
"""

import unittest
import numpy as np

import andes
import ams

from ams.interface import (build_group_table, make_link_table, to_andes,
                           parse_addfile, verify_pf)


class TestAndesConversion(unittest.TestCase):
    """
    Tests conversion from AMS to ANDES.
    """
    ad_cases = [
        'ieee14/ieee14_full.xlsx',
        'ieee39/ieee39_full.xlsx',
        'npcc/npcc.xlsx',
    ]
    am_cases = [
        'ieee14/ieee14.json',
        'ieee39/ieee39.xlsx',
        'npcc/npcc_uced.xlsx',
    ]

    def setUp(self) -> None:
        """
        Test setup.
        """

    def test_basic_functions(self):
        """
        Test basic functions defined in ANDES Interop.
        """
        for ad_case in self.ad_cases:
            sa = andes.load(andes.get_case(ad_case),
                            setup=True, no_output=True, default_config=True,)
            # --- test build_group_table ---
            ssa_stg = build_group_table(adsys=sa,
                                        grp_name='StaticGen',
                                        param_name=['u', 'name', 'idx', 'bus'],
                                        mdl_name=[])
            self.assertEqual(ssa_stg.shape,
                             (sa.StaticGen.n, 4))

            ssa_gov = build_group_table(adsys=sa, grp_name='TurbineGov',
                                        param_name=['idx', 'syn'],
                                        mdl_name=[])
            self.assertEqual(ssa_gov.shape,
                             (sa.TurbineGov.n, 2))
            # --- test make_link_table ---
            link_table = make_link_table(adsys=sa)
            stg_idx = [str(i) for i in sa.PV.idx.v + sa.Slack.idx.v]
            self.assertSetEqual(set(stg_idx),
                                set(link_table['stg_idx'].values))
            bus_idx = [str(i) for i in sa.PV.bus.v + sa.Slack.bus.v]
            self.assertSetEqual(set(bus_idx),
                                set(link_table['bus_idx'].values))

    def test_convert(self):
        """
        Test conversion from AMS case to ANDES case.
        """
        for ad_case, am_case in zip(self.ad_cases, self.am_cases):
            sp = ams.load(ams.get_case(am_case),
                          setup=True, no_output=True, default_config=True,)
            # before addfile
            sa = to_andes(sp, setup=False)
            self.assertEqual(set(sp.PV.idx.v), set(sa.PV.idx.v))
            self.assertEqual(set(sp.Bus.idx.v), set(sa.Bus.idx.v))
            self.assertEqual(set(sp.Line.idx.v), set(sa.Line.idx.v))
            self.assertEqual(np.sum(sp.PQ.p0.v), np.sum(sa.PQ.p0.v))

            # after addfile
            sa = parse_addfile(adsys=sa, amsys=sp,
                               addfile=andes.get_case(ad_case))
            sa.setup()
            set1 = set(sa.GENROU.gen.v)
            set2 = set(sp.StaticGen.get_all_idxes())
            # set2 includes set1, ensure GENROU.gen are all in StaticGen.idx
            self.assertEqual(set1, set1 & set2)

            # ensure PFlow models consistency
            pflow_mdls = list(sa.PFlow.models.keys())
            for mdl in pflow_mdls:
                self.assertTrue(sp.models[mdl].as_df().equals(sa.PFlow.models[mdl].as_df()))

    def test_convert_after_update(self):
        """
        Test conversion from AMS case to ANDES case after updating parameters.
        """
        for am_case in self.am_cases:
            sp = ams.load(ams.get_case(am_case),
                          setup=True, no_output=True, default_config=True,)
        # record initial values
        pq_idx = sp.PQ.idx.v
        p0 = sp.PQ.p0.v.copy()
        sa = to_andes(sp, setup=False, no_output=True, default_config=True)
        # before update
        np.testing.assert_array_equal(sp.PQ.p0.v, sa.PQ.p0.v)
        # after update
        sp.PQ.alter(src='p0', idx=pq_idx, value=0.9*p0)
        sa = to_andes(sp, setup=False, no_output=True, default_config=True)
        np.testing.assert_array_equal(sp.PQ.p0.v, sa.PQ.p0.v)

    def test_extra_dyn(self):
        """
        Test conversion when extra dynamic models exist.
        """
        sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'),
                      setup=True, no_output=True, default_config=True,)
        sa = to_andes(sp, addfile=andes.get_case('ieee14/ieee14_full.xlsx'),
                      setup=True, no_output=True, default_config=True,
                      verify=False, tol=1e-3)
        self.assertGreaterEqual(sa.PVD1.n, 0)

    def test_verify_pf(self):
        """
        Test verification of power flow results.
        """
        sp = ams.load(ams.get_case('matpower/case300.m'),
                      setup=True, no_output=True, default_config=True,)
        sa = to_andes(sp,
                      setup=True, no_output=True, default_config=True,
                      verify=False, tol=1e-3)
        self.assertTrue(verify_pf(amsys=sp, adsys=sa, tol=1e-7))

    def test_to_andes_toggle(self):
        """
        Test conversion when there is Toggle in ANDES case.
        """
        sp = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True,
                      no_output=True,
                      default_config=True)
        sa = sp.to_andes(setup=True,
                         addfile=andes.get_case('5bus/pjm5bus.xlsx'),
                         verify=False)
        self.assertGreater(sa.Toggle.n, 0)

    def test_to_andes_alter(self):
        """
        Test conversion when there is Alter in ANDES case.
        """
        sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'),
                      setup=True,
                      no_output=True,
                      default_config=True)
        sa = sp.to_andes(setup=True,
                         addfile=andes.get_case('ieee14/ieee14_alter.xlsx'),
                         verify=False)
        self.assertGreater(sa.Alter.n, 0)

    def test_to_andes_fault(self):
        """
        Test conversion when there is Toggle in ANDES case.
        """
        sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'),
                      setup=True,
                      no_output=True,
                      default_config=True)
        sa = sp.to_andes(setup=True,
                         addfile=andes.get_case('ieee14/ieee14_fault.xlsx'),
                         verify=False)
        self.assertGreater(sa.Fault.n, 0)


class TestDataExchange(unittest.TestCase):
    """
    Tests for data exchange between AMS and ANDES.
    """

    def setUp(self) -> None:
        """
        Test setup. This is executed before each test case.
        """
        self.sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'),
                           setup=True,
                           no_output=True,
                           default_config=True,)
        self.sp.RTED.run(solver='CLARABEL')
        self.sp.RTED.dc2ac()
        self.stg_idx = self.sp.RTED.pg.get_all_idxes()

    def test_data_exchange(self):
        """
        Test data exchange between AMS and ANDES.
        """
        sa = to_andes(self.sp, setup=True,
                      addfile=andes.get_case('ieee14/ieee14_full.xlsx'),
                      no_output=True,
                      default_config=True,)
        # alleviate limiter
        sa.TGOV1.set(src='VMAX', idx=sa.TGOV1.idx.v, attr='v', value=100*np.ones(sa.TGOV1.n))
        sa.TGOV1.set(src='VMIN', idx=sa.TGOV1.idx.v, attr='v', value=np.zeros(sa.TGOV1.n))

        # --- test before PFlow ---
        self.sp.dyn.send(adsys=sa, routine='RTED')
        p0 = sa.StaticGen.get(src='p0', attr='v', idx=self.stg_idx)
        pg = self.sp.RTED.get(src='pg', attr='v', idx=self.stg_idx)
        np.testing.assert_array_equal(p0, pg)

        # --- test after TDS ---
        self.assertFalse(self.sp.dyn.is_tds)
        sa.PFlow.run()
        sa.TDS.init()
        self.assertTrue(self.sp.dyn.is_tds)

        sa.TDS.config.tf = 1
        sa.TDS.run()
        self.sp.dyn.send(adsys=sa, routine='RTED')
        self.sp.dyn.receive(adsys=sa, routine='RTED', no_update=False)



from andes.models.line.line import LineData
from andes.models.line.jumper import JumperData
from andes.core.param import NumParam
from andes.shared import deg2rad, np, spmatrix

from ams.core.model import Model


class Line(LineData, Model):
    """
    AC transmission line model.

    The model is also used for two-winding transformer. Transformers can set the
    tap ratio in ``tap`` and/or phase shift angle ``phi``.

    Note that the bus admittance matrix is built on fly and is not stored in the
    object.

    Notes
    -----
    There is a known issue that adding Algeb ``ud`` will cause Line.algebs run into
    AttributeError: 'NoneType' object has no attribute 'n'. Not figured out why yet.
    """

    def __init__(self, system=None, config=None) -> None:
        LineData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACLine'

        self.amin = NumParam(default=-360 * deg2rad,
                             info="minimum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{min}',
                             )
        self.amax = NumParam(default=360 * deg2rad,
                             info="maximum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{max}',
                             )
        self.rate_a.unit = 'p.u.'
        self.rate_b.unit = 'p.u.'
        self.rate_c.unit = 'p.u.'
        self.rate_a.default = 999.0
        self.rate_b.default = 999.0
        self.rate_c.default = 999.0

        # NOTE: following parameters are prepared for building matrices
        # they are initialized here but populated in ``System.setup()``.
        self.a1a = None
        self.a2a = None

    # NOTE: following code are minly copied from `andes.models.line.Line`
    # and they are not fully verified
    # potential issues:
    # `build_Bp` contains 'fdxb', which is not included in the input parameters,
    # and the results are the negative of `Bbus` from `makeBdc` in PYPOWER
    # `build_Bpp` ignores the line resistance for all three methods
    # `build_Bdc` results are the negative of `Bbus` from `makeBdc` in PYPOWER
    # `build_y` results have inignorable differences at diagonal elements with `makeYbus` in PYPOWER

    def build_y(self):
        """
        Build bus admittance matrix. Copied from ``andes.models.line.line.Line``.

        Returns
        -------
        Y : spmatrix
            Bus admittance matrix.
        """

        nb = self.system.Bus.n

        y1 = self.u.v * (self.g1.v + self.b1.v * 1j)
        y2 = self.u.v * (self.g2.v + self.b2.v * 1j)
        y12 = self.u.v / (self.r.v + self.x.v * 1j)
        m = self.tap.v * np.exp(1j * self.phi.v)
        m2 = self.tap.v**2
        mconj = np.conj(m)

        # build self and mutual admittances into Y
        Y = spmatrix((y12 + y1 / m2), self.a1a, self.a1a, (nb, nb), 'z')
        Y -= spmatrix(y12 / mconj, self.a1a, self.a2a, (nb, nb), 'z')
        Y -= spmatrix(y12 / m, self.a2a, self.a1a, (nb, nb), 'z')
        Y += spmatrix(y12 + y2, self.a2a, self.a2a, (nb, nb), 'z')

        return Y

    def build_Bp(self, method='fdpf'):
        """
        Function for building B' matrix.

        Parameters
        ----------
        method : str
            Method for building B' matrix. Choose from 'fdpf', 'fdbx', 'dcpf'.

        Returns
        -------
        Bp : spmatrix
            B' matrix.
        """
        nb = self.system.Bus.n

        if method not in ("fdpf", "fdbx", "dcpf"):
            raise ValueError(f"Invalid method {method}; choose from 'fdpf', 'fdbx', 'dcpf'")

        # Build B prime matrix -- FDPF
        # `y1`` neglects line charging shunt, and g1 is usually 0 in HV lines
        # `y2`` neglects line charging shunt, and g2 is usually 0 in HV lines
        y1 = self.u.v * self.g1.v
        y2 = self.u.v * self.g2.v

        # `m` neglected tap ratio
        m = np.exp(self.phi.v * 1j)
        mconj = np.conj(m)
        m2 = np.ones(self.n)

        if method in ('fdxb', 'dcpf'):
            # neglect line resistance in Bp in XB method
            y12 = self.u.v / (self.x.v * 1j)
        else:
            y12 = self.u.v / (self.r.v + self.x.v * 1j)

        Bdc = spmatrix((y12 + y1) / m2, self.a1a, self.a1a, (nb, nb), 'z')
        Bdc -= spmatrix(y12 / mconj, self.a1a, self.a2a, (nb, nb), 'z')
        Bdc -= spmatrix(y12 / m, self.a2a, self.a1a, (nb, nb), 'z')
        Bdc += spmatrix(y12 + y2, self.a2a, self.a2a, (nb, nb), 'z')
        Bdc = Bdc.imag()

        for item in range(nb):
            if abs(Bdc[item, item]) == 0:
                Bdc[item, item] = 1e-6 + 0j

        return Bdc

    def build_Bpp(self, method='fdpf'):
        """
        Function for building B'' matrix.

        Parameters
        ----------
        method : str
            Method for building B'' matrix. Choose from 'fdpf', 'fdbx', 'dcpf'.

        Returns
        -------
        Bpp : spmatrix
            B'' matrix.
        """

        nb = self.system.Bus.n

        if method not in ("fdpf", "fdbx", "dcpf"):
            raise ValueError(f"Invalid method {method}; choose from 'fdpf', 'fdbx', 'dcpf'")

        # Build B double prime matrix
        # y1 neglected line charging shunt, and g1 is usually 0 in HV lines
        # y2 neglected line charging shunt, and g2 is usually 0 in HV lines
        # m neglected phase shifter
        y1 = self.u.v * (self.g1.v + self.b1.v * 1j)
        y2 = self.u.v * (self.g2.v + self.b2.v * 1j)

        m = self.tap.v
        m2 = abs(m)**2

        if method in ('fdbx', 'fdpf', 'dcpf'):
            # neglect line resistance in Bpp in BX method
            y12 = self.u.v / (self.x.v * 1j)
        else:
            y12 = self.u.v / (self.r.v + self.x.v * 1j)

        Bpp = spmatrix((y12 + y1) / m2, self.a1a, self.a1a, (nb, nb), 'z')
        Bpp -= spmatrix(y12 / np.conj(m), self.a1a, self.a2a, (nb, nb), 'z')
        Bpp -= spmatrix(y12 / m, self.a2a, self.a1a, (nb, nb), 'z')
        Bpp += spmatrix(y12 + y2, self.a2a, self.a2a, (nb, nb), 'z')
        Bpp = Bpp.imag()

        for item in range(nb):
            if abs(Bpp[item, item]) == 0:
                Bpp[item, item] = 1e-6 + 0j

        return Bpp

    def build_Bdc(self):
        """
        The MATPOWER-flavor Bdc matrix for DC power flow.

        The method neglects line charging and line resistance. It retains tap ratio.

        Returns
        -------
        Bdc : spmatrix
            Bdc matrix.
        """

        nb = self.system.Bus.n

        y12 = self.u.v / (self.x.v * 1j)
        y12 = y12 / self.tap.v

        Bdc = spmatrix(y12, self.a1a, self.a1a, (nb, nb), 'z')
        Bdc -= spmatrix(y12, self.a1a, self.a2a, (nb, nb), 'z')
        Bdc -= spmatrix(y12, self.a2a, self.a1a, (nb, nb), 'z')
        Bdc += spmatrix(y12, self.a2a, self.a2a, (nb, nb), 'z')
        Bdc = Bdc.imag()

        for item in range(nb):
            if abs(Bdc[item, item]) == 0:
                Bdc[item, item] = 1e-6

        return Bdc


class Jumper(JumperData, Model):
    """
    Jumper is a device to short two buses (merging two buses into one).

    Jumper can connect two buses satisfying one of the following conditions:

    - neither bus is voltage-controlled
    - either bus is voltage-controlled
    - both buses are voltage-controlled, and the voltages are the same.

    If the buses are controlled in different voltages, power flow will
    not solve (as the power flow through the jumper will be infinite).

    In the solutions, the ``p`` and ``q`` are flowing out of bus1
    and flowing into bus2.

    Setting a Jumper's connectivity status ``u`` to zero will disconnect the two
    buses. In the case of a system split, one will need to call
    ``System.connectivity()`` immediately following the split to detect islands.
    """

    def __init__(self, system=None, config=None) -> None:
        JumperData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACShort'



"""
Module for optimization OModel.
"""
import logging

from typing import Any
from collections import OrderedDict

from andes.utils.misc import elapsed

import cvxpy as cp

from ams.opt.optzbase import ensure_symbols, ensure_mats_and_parsed


logger = logging.getLogger(__name__)


class OModelBase:
    """
    Template class for optimization models.
    """

    def __init__(self, routine):
        self.rtn = routine
        self.prob = None
        self.exprs = OrderedDict()
        self.params = OrderedDict()
        self.vars = OrderedDict()
        self.constrs = OrderedDict()
        self.obj = None
        self.parsed = False
        self.evaluated = False
        self.finalized = False

    @property
    def initialized(self):
        """
        Return the initialization status.
        """
        return self.parsed and self.evaluated and self.finalized

    def parse(self, force=False):
        self.parsed = True
        return self.parsed

    def _evaluate_params(self):
        return True

    def _evaluate_vars(self):
        return True

    def _evaluate_constrs(self):
        return True

    def _evaluate_obj(self):
        return True

    def _evaluate_exprs(self):
        return True

    def _evaluate_exprcs(self):
        return True

    def evaluate(self, force=False):
        self._evaluate_params()
        self._evaluate_vars()
        self._evaluate_exprs()
        self._evaluate_constrs()
        self._evaluate_obj()
        self._evaluate_exprcs()
        self.evaluated = True
        return self.evaluated

    def finalize(self, force=False):
        self.finalized = True
        return True

    def init(self, force=False):
        self.parse(force)
        self.evaluate(force)
        self.finalize(force)
        return self.initialized

    @property
    def class_name(self):
        return self.__class__.__name__

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to OModel instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        """
        if isinstance(value, cp.Variable):
            self.vars[key] = value
        elif isinstance(value, cp.Constraint):
            self.constrs[key] = value
        elif isinstance(value, cp.Parameter):
            self.params[key] = value
        elif isinstance(value, cp.Expression):
            self.exprs[key] = value

    def __setattr__(self, name: str, value: Any):
        super().__setattr__(name, value)
        self._register_attribute(name, value)

    def update(self, params):
        return True

    def __repr__(self) -> str:
        return f'{self.rtn.class_name}.{self.__class__.__name__} at {hex(id(self))}'


class OModel(OModelBase):
    """
    Base class for optimization models.

    Parameters
    ----------
    routine: Routine
        Routine that to be modeled.

    Attributes
    ----------
    prob: cvxpy.Problem
        Optimization model.
    exprs: OrderedDict
        Expressions registry.
    params: OrderedDict
        Parameters registry.
    vars: OrderedDict
        Decision variables registry.
    constrs: OrderedDict
        Constraints registry.
    obj: Objective
        Objective function.
    initialized: bool
        Flag indicating if the model is initialized.
    parsed: bool
        Flag indicating if the model is parsed.
    evaluated: bool
        Flag indicating if the model is evaluated.
    finalized: bool
        Flag indicating if the model is finalized.
    """

    def __init__(self, routine):
        OModelBase.__init__(self, routine)

    @ensure_symbols
    def parse(self, force=False):
        """
        Parse the optimization model from the symbolic description.

        This method should be called after the routine symbols are generated
        `self.rtn.syms.generate_symbols()`. It parses the following components
        of the optimization model: parameters, decision variables, constraints,
        objective function, and expressions.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the parsing.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        if self.parsed and not force:
            logger.debug("Model is already parsed.")
            return self.parsed

        t, _ = elapsed()
        logger.warning(f'Parsing OModel for <{self.rtn.class_name}>')
        # --- add expressions ---
        for key, val in self.rtn.exprs.items():
            val.parse()

        # --- add RParams and Services as parameters ---
        for key, val in self.rtn.params.items():
            if not val.no_parse:
                val.parse()

        # --- add decision variables ---
        for key, val in self.rtn.vars.items():
            val.parse()

        # --- add constraints ---
        for key, val in self.rtn.constrs.items():
            val.parse()

        # --- add ExpressionCalcs ---
        for key, val in self.rtn.exprcs.items():
            val.parse()

        # --- parse objective functions ---
        if self.rtn.obj is not None:
            try:
                self.rtn.obj.parse()
            except Exception as e:
                raise Exception(f"Failed to parse Objective <{self.rtn.obj.name}>.\n{e}")
        elif self.rtn.class_name not in ['DCPF0']:
            logger.warning(f"{self.rtn.class_name} has no objective function!")
            self.parsed = False
            return self.parsed

        # --- parse expressions ---
        for key, val in self.rtn.exprs.items():
            try:
                val.parse()
            except Exception as e:
                raise Exception(f"Failed to parse ExpressionCalc <{key}>.\n{e}")
        _, s = elapsed(t)
        logger.debug(f"  -> Parsed in {s}")

        self.parsed = True
        return self.parsed

    def _evaluate_params(self):
        """
        Evaluate the parameters.
        """
        for key, val in self.rtn.params.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Param <{key}>.\n{e}")

    def _evaluate_vars(self):
        """
        Evaluate the decision variables.
        """
        for key, val in self.rtn.vars.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Var <{key}>.\n{e}")

    def _evaluate_constrs(self):
        """
        Evaluate the constraints.
        """
        for key, val in self.rtn.constrs.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Constr <{key}>.\n{e}")

    def _evaluate_obj(self):
        """
        Evaluate the objective function.
        """
        # NOTE: since we already have the attribute `obj`,
        # we can update it rather than setting it
        if self.rtn.obj is not None:
            self.rtn.obj.evaluate()
            self.obj = self.rtn.obj.optz

    def _evaluate_exprs(self):
        """
        Evaluate the expressions.
        """
        for key, val in self.rtn.exprs.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Expression <{key}>.\n{e}")

    def _evaluate_exprcs(self):
        """
        Evaluate the expressions.
        """
        for key, val in self.rtn.exprcs.items():
            try:
                val.evaluate()
            except Exception as e:
                raise Exception(f"Failed to evaluate ExpressionCalc <{key}>.\n{e}")

    @ensure_mats_and_parsed
    def evaluate(self, force=False):
        """
        Evaluate the optimization model.

        This method should be called after `self.parse()`. It evaluates the following
        components of the optimization model: parameters, decision variables, constraints,
        objective function, and expressions.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the evaluation

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        if self.evaluated and not force:
            logger.debug("Model is already evaluated.")
            return self.evaluated
        logger.warning(f"Evaluating OModel for <{self.rtn.class_name}>")
        t, _ = elapsed()

        # NOTE: should evaluate in sequence
        self._evaluate_params()
        self._evaluate_vars()
        self._evaluate_exprs()
        self._evaluate_constrs()
        self._evaluate_obj()
        self._evaluate_exprcs()

        self.evaluated = True
        _, s = elapsed(t)
        logger.debug(f" -> Evaluated in {s}")
        return self.evaluated

    def finalize(self, force=False):
        """
        Finalize the optimization model.

        This method should be called after `self.evaluate()`. It assemble the optimization
        problem from the evaluated components.

        Returns
        -------
        bool
            Returns True if the finalization is successful, False otherwise.
        """
        # NOTE: for power flow type, we skip the finalization
        if self.rtn.class_name in ['DCPF0']:
            self.finalized = True
            return self.finalized
        if self.finalized and not force:
            logger.debug("Model is already finalized.")
            return self.finalized
        logger.warning(f"Finalizing OModel for <{self.rtn.class_name}>")
        t, _ = elapsed()

        # Collect constraints that are not disabled
        constrs_add = [val.optz for key, val in self.rtn.constrs.items(
        ) if not val.is_disabled and val is not None]
        # Construct the problem using cvxpy.Problem
        self.prob = cp.Problem(self.obj, constrs_add)

        _, s = elapsed(t)
        logger.debug(f" -> Finalized in {s}")
        self.finalized = True
        return self.finalized

    def init(self, force=False):
        """
        Set up the optimization model from the symbolic description.

        This method initializes the optimization model by parsing decision variables,
        constraints, and the objective function from the associated routine.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the OModel initialization.
            If True, following methods will be called by force: `self.parse()`,
            `self.evaluate()`, `self.finalize()`

        Returns
        -------
        bool
            Returns True if the setup is successful, False otherwise.
        """
        if self.initialized and not force:
            logger.debug("OModel is already initialized.")
            return self.initialized

        t, _ = elapsed()

        self.parse(force=force)
        self.evaluate(force=force)
        self.finalize(force=force)

        _, s = elapsed(t)
        logger.debug(f"OModel for <{self.rtn.class_name}> initialized in {s}")

        return self.initialized

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to OModel instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        """
        if isinstance(value, cp.Variable):
            self.vars[key] = value
        elif isinstance(value, cp.Constraint):
            self.constrs[key] = value
        elif isinstance(value, cp.Parameter):
            self.params[key] = value
        elif isinstance(value, cp.Expression):
            self.exprs[key] = value

    def __setattr__(self, name: str, value: Any):
        super().__setattr__(name, value)
        self._register_attribute(name, value)

    def update(self, params):
        """
        Update the Parameter values.

        Parameters
        ----------
        params: list
            List of parameters to be updated.
        """
        for param in params:
            param.update()
        return True

    def __repr__(self) -> str:
        return f'{self.rtn.class_name}.{self.__class__.__name__} at {hex(id(self))}'



"""
Module for optimization Constraint.
"""
import logging

from typing import Optional
import re

import numpy as np

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import _prefix, _max_length

from ams.opt import OptzBase, ensure_symbols, ensure_mats_and_parsed

logger = logging.getLogger(__name__)


class Constraint(OptzBase):
    """
    Base class for constraints.

    This class is used as a template for defining constraints. Each
    instance of this class represents a single constraint.

    Parameters
    ----------
    name : str, optional
        A user-defined name for the constraint.
    e_str : str, optional
        A mathematical expression representing the constraint.
    info : str, optional
        Additional informational text about the constraint.
    is_eq : str, optional
        Flag indicating if the constraint is an equality constraint. False indicates
        an inequality constraint in the form of `<= 0`.

    Attributes
    ----------
    is_disabled : bool
        Flag indicating if the constraint is disabled, False by default.
    rtn : ams.routines.Routine
        The owner routine instance.
    is_disabled : bool, optional
        Flag indicating if the constraint is disabled, False by default.
    code : str, optional
        The code string for the constraint
    """

    def __init__(self,
                 name: Optional[str] = None,
                 e_str: Optional[str] = None,
                 info: Optional[str] = None,
                 is_eq: Optional[bool] = False,
                 ):
        OptzBase.__init__(self, name=name, info=info)
        self.e_str = e_str
        self.is_eq = is_eq
        self.is_disabled = False
        self.code = None

    def get_idx(self):
        raise NotImplementedError

    def get_all_idxes(self):
        raise NotImplementedError

    @ensure_symbols
    def parse(self):
        """
        Parse the constraint.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_constr = self.e_str
        for pattern, replacement in sub_map.items():
            try:
                code_constr = re.sub(pattern, replacement, code_constr)
            except TypeError as e:
                raise TypeError(f"Error in parsing constr <{self.name}>.\n{e}")
        # parse the constraint type
        code_constr += " == 0" if self.is_eq else " <= 0"
        # store the parsed expression str code
        self.code = code_constr
        msg = f" - Constr <{self.name}>: {self.e_str}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the constraint.
        """
        msg = f" - Constr <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'cp': cp, 'sub_map': self.om.rtn.syms.val_map}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating Constraint <{self.name}>.\n{e}")

    def __repr__(self):
        enabled = 'OFF' if self.is_disabled else 'ON'
        out = f"{self.class_name}: {self.name} [{enabled}]"
        return out

    @property
    def e(self):
        """
        Return the calculated constraint LHS value.
        Note that `v` should be used primarily as it is obtained
        from the solver directly.

        `e` is for debugging purpose. For a successfully solved problem,
        `e` should equal to `v`. However, when a problem is infeasible
        or unbounded, `e` can be used to check the constraint LHS value.
        """
        if self.code is None:
            logger.info(f"Constraint <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                raise TypeError(e)

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating constr <{self.name}>.\n{e}")
            return None

    @property
    def v(self):
        """
        Return the CVXPY constraint LHS value.
        """
        if self.optz is None:
            return None
        if self.optz._expr.value is None:
            try:
                shape = self._expr.shape
                return np.zeros(shape)
            except AttributeError:
                return None
        else:
            return self.optz._expr.value

    @v.setter
    def v(self, value):
        raise AttributeError("Cannot set the value of the constraint.")



"""
Runs a full AC continuation power flow.
"""
# --- run cpf ---
import logging

import numpy as np
from numpy import flatnonzero as find

import scipy.sparse as sp

from andes.shared import deg2rad
from andes.utils.misc import elapsed

import ams.pypower.utils as putil
import ams.pypower.io as pio
import ams.pypower.routines.opffcns as opfcn
from ams.pypower.idx import IDX
from ams.pypower.make import (makeSbus, makeYbus, dSbus_dV)
from ams.pypower.routines.pflow import newtonpf, pfsoln
from ams.pypower.core import ppoption
import ams.pypower.routines.cpf_callbacks as cpf_callbacks

from ams.shared import inf, nan

logger = logging.getLogger(__name__)


def runcpf(casedata, ppopt=None, scale=1.2):
    """
    Runs a full AC continuation power flow.
    """
    ppopt = ppoption(ppopt)

    # options
    verbose = ppopt["VERBOSE"]
    step = ppopt["CPF_STEP"]
    parameterization = ppopt["CPF_PARAMETERIZATION"]
    adapt_step = ppopt["CPF_ADAPT_STEP"]
    cb_args = ppopt["CPF_USER_CALLBACK_ARGS"]

    # set up callbacks
    callback_names = ["cpf_default_callback"]
    if len(ppopt["CPF_USER_CALLBACK"]) > 0:
        if isinstance(ppopt["CPF_USER_CALLBACK"], list):
            callback_names = np.r_[callback_names, ppopt["CPF_USER_CALLBACK"]]
        else:
            callback_names.append(ppopt["CPF_USER_CALLBACK"])
    callbacks = []
    for callback_name in callback_names:
        callbacks.append(getattr(cpf_callbacks, callback_name))

    # read base case data
    ppcbase = pio.loadcase(casedata)
    nb = ppcbase["bus"].shape[0]

    # add zero columns to branch for flows if needed
    if ppcbase["branch"].shape[1] < IDX.branch.QT:
        ppcbase["branch"] = np.c_[ppcbase["branch"],
                                  np.zeros((ppcbase["branch"].shape[0],
                                            IDX.branch.QT - ppcbase["branch"].shape[1] + 1))]

    # convert to internal indexing
    ppcbase = opfcn.ext2int(ppcbase)
    baseMVAb, busb, genb, branchb = \
        ppcbase["baseMVA"], ppcbase["bus"], ppcbase["gen"], ppcbase["branch"]

    # get bus index lists of each type of bus
    ref, pv, pq = putil.bustypes(busb, genb)

    # generator info
    onb = find(genb[:, IDX.gen.GEN_STATUS] > 0)  # which generators are on?
    gbusb = genb[onb, IDX.gen.GEN_BUS].astype(int)  # what buses are they at?

    # scale the load and generation of base case as target case
    ppctarget = pio.loadcase(casedata)
    ppctarget["bus"][:, IDX.bus.PD] *= scale
    ppctarget["bus"][:, IDX.bus.QD] *= scale
    ppctarget["gen"][:, IDX.gen.PG] *= scale
    ppctarget["gen"][:, IDX.gen.QG] *= scale

    # add zero columns to branch for flows if needed
    if ppctarget["branch"].shape[1] < IDX.branch.QT:
        ppctarget["branch"] = np.c_[ppctarget["branch"],
                                    np.zeros((ppctarget["branch"].shape[0],
                                              IDX.branch.QT - ppctarget["branch"].shape[1] + 1))]

    # convert to internal indexing
    ppctarget = opfcn.ext2int(ppctarget)
    baseMVAt, bust, gent, brancht = \
        ppctarget["baseMVA"], ppctarget["bus"], ppctarget["gen"], ppctarget["branch"]

    # get bus index lists of each type of bus
    # ref, pv, pq = putil.bustypes(bust, gent)

    # generator info
    ont = find(gent[:, IDX.gen.GEN_STATUS] > 0)  # which generators are on?
    # gbust = gent[ont, IDX.gen.GEN_BUS].astype(int)  # what buses are they at?

    # -----  run the power flow  -----
    t0, _ = elapsed()  # start timer

    # initial state
    # V0    = np.ones((bus.shape[0])            ## flat start
    V0 = busb[:, IDX.bus.VM] * np.exp((1j * deg2rad * busb[:, IDX.bus.VA]))
    vcb = np.ones(V0.shape)    # create mask of voltage-controlled buses
    vcb[pq] = 0     # exclude PQ buses
    k = find(vcb[gbusb])    # in-service gens at v-c buses
    V0[gbusb[k]] = genb[onb[k], IDX.gen.VG] / abs(V0[gbusb[k]]) * V0[gbusb[k]]

    # build admittance matrices
    Ybus, Yf, Yt = makeYbus(baseMVAb, busb, branchb)

    # compute base case complex bus power injections (generation - load)
    Sbusb = makeSbus(baseMVAb, busb, genb)
    # compute target case complex bus power injections (generation - load)
    Sbust = makeSbus(baseMVAt, bust, gent)

    # scheduled transfer
    Sxfr = Sbust - Sbusb

    # Run the base case power flow solution
    lam = 0
    logger.info('Solve base case power flow')
    V, success, iterations = newtonpf(Ybus, Sbusb, V0, ref, pv, pq, ppopt)

    logger.info('Start full AC CPF')
    logger.info('%2d: lambda = %6.3f, %2d Newton steps' % (0, 0, iterations))

    lamprv = lam    # lam at previous step
    Vprv = V    # V at previous step
    continuation = 1
    cont_steps = 0

    # input args for callbacks
    cb_data = dict(ppc_base=ppcbase, ppc_target=ppctarget,
                   Sxfr=Sxfr, Ybus=Ybus, Yf=Yf, Yt=Yt,
                   ref=ref, pv=pv, pq=pq, ppopt=ppopt)
    cb_state = {}

    # invoke callbacks
    for k in range(len(callbacks)):
        cb_state, _ = callbacks[k](cont_steps, V, lam, V, lam,
                                   cb_data, cb_state, cb_args)

    if np.linalg.norm(Sxfr) == 0:
        logger.error('base case and target case have identical load and generation!')

        continuation = 0
        V0 = V
        lam0 = lam

    # tangent predictor z = [dx;dlam]
    z = np.zeros(2*len(V)+1)
    z[-1] = 1.0
    while continuation:
        cont_steps = cont_steps + 1
        # prediction for next step
        V0, lam0, z = cpf_predictor(V, lam, Ybus, Sxfr, pv, pq, step, z,
                                    Vprv, lamprv, parameterization)

        # save previous voltage, lambda before updating
        Vprv = V
        lamprv = lam

        # correction
        V, success, i, lam = cpf_corrector(Ybus, Sbusb, V0, ref, pv, pq,
                                           lam0, Sxfr, Vprv, lamprv, z,
                                           step, parameterization, ppopt)

        if not success:
            continuation = 0
            logger.warning('%2d: lambda = %6.3f, corrector did not converge in %d iterations' % (
                cont_steps, lam, i))
            break

        logger.info('%2d: lambda = %6.3f, %2d corrector Newton steps' %
                    (cont_steps, lam, i))

        # invoke callbacks
        for k in range(len(callbacks)):
            cb_state, _ = callbacks[k](cont_steps, V, lam, V0, lam0,
                                       cb_data, cb_state, cb_args)

        if isinstance(ppopt["CPF_STOP_AT"], str):
            if ppopt["CPF_STOP_AT"].upper() == "FULL":
                if abs(lam) < 1e-8:     # traced the full continuation curve
                    logger.info('Reached steady state loading limit in %d continuation steps' % cont_steps)
                    continuation = 0
                elif lam < lamprv and lam - step < 0:    # next step will overshoot
                    step = lam      # modify step-size
                    parameterization = 1    # change to natural parameterization
                    adapt_step = False      # disable step-adaptivity

            else:   # == 'NOSE'
                if lam < lamprv:    # reached the nose point
                    logger.info('Reached steady state loading limit in %d continuation steps' % cont_steps)
                    continuation = 0

        else:
            if lam < lamprv:
                logger.info('Reached steady state loading limit in %d continuation steps' % cont_steps)
                continuation = 0
            elif abs(ppopt["CPF_STOP_AT"] - lam) < 1e-8:     # reached desired lambda
                logger.info('Reached desired lambda %3.2f in %d continuation steps' % (
                    ppopt["CPF_STOP_AT"], cont_steps))
                continuation = 0
            # will reach desired lambda in next step
            elif lam + step > ppopt["CPF_STOP_AT"]:
                step = ppopt["CPF_STOP_AT"] - lam   # modify step-size
                parameterization = 1    # change to natural parameterization
                adapt_step = False      # disable step-adaptivity

        if adapt_step and continuation:
            pvpq = np.r_[pv, pq]
            # Adapt stepsize
            cpf_error = np.linalg.norm(np.r_[np.angle(V[pq]), abs(
                V[pvpq]), lam] - np.r_[np.angle(V0[pq]), abs(V0[pvpq]), lam0], inf)
            if cpf_error < ppopt["CPF_ERROR_TOL"]:
                # Increase stepsize
                step = step * ppopt["CPF_ERROR_TOL"] / cpf_error
                if step > ppopt["CPF_STEP_MAX"]:
                    step = ppopt["CPF_STEP_MAX"]
            else:
                # decrese stepsize
                step = step * ppopt["CPF_ERROR_TOL"] / cpf_error
                if step < ppopt["CPF_STEP_MIN"]:
                    step = ppopt["CPF_STEP_MIN"]

    # invoke callbacks
    if success:
        cpf_results = {}
        for k in range(len(callbacks)):
            cb_state, cpf_results = callbacks[k](cont_steps, V, lam, V0, lam0,
                                                 cb_data, cb_state, cb_args, results=cpf_results, is_final=True)
    else:
        cpf_results = {}
        cpf_results["iterations"] = i

    # update bus and gen matrices to reflect the loading and generation
    # at the noise point
    bust[:, IDX.bus.PD] = busb[:, IDX.bus.PD] + lam * (bust[:, IDX.bus.PD] - busb[:, IDX.bus.PD])
    bust[:, IDX.bus.QD] = busb[:, IDX.bus.QD] + lam * (bust[:, IDX.bus.QD] - busb[:, IDX.bus.QD])
    gent[:, IDX.gen.PG] = genb[:, IDX.gen.PG] + lam * (gent[:, IDX.gen.PG] - genb[:, IDX.gen.PG])

    # update data matrices with solution
    bust, gent, brancht = pfsoln(baseMVAt, bust, gent, brancht,
                                 Ybus, Yf, Yt, V, ref, pv, pq)

    ppctarget["et"] = elapsed(t0)
    ppctarget["success"] = success

    # -----  output results  -----
    # convert back to original bus numbering & print results
    ppctarget["bus"], ppctarget["gen"], ppctarget["branch"] = bust, gent, brancht
    if success:
        n = cpf_results["iterations"] + 1
        cpf_results["V_p"] = opfcn.i2e_data(
            ppctarget, cpf_results["V_p"], np.full((nb, n), nan), "bus", 0)
        cpf_results["V_c"] = opfcn.i2e_data(
            ppctarget, cpf_results["V_c"], np.full((nb, n), nan), "bus", 0)
    results = opfcn.int2ext(ppctarget)
    results["cpf"] = cpf_results

    # zero out result fields of out-of-service gens & branches
    if len(results["order"]["gen"]["status"]["off"]) > 0:
        results["gen"][np.ix_(results["order"]["gen"]
                              ["status"]["off"], [IDX.gen.QF, IDX.gen.QG])] = 0

    if len(results["order"]["branch"]["status"]["off"]) > 0:
        results["branch"][np.ix_(results["order"]["branch"]
                                 ["status"]["off"], [IDX.branch.PF, IDX.branch.QF, IDX.branch.PT, IDX.branch.QT])] = 0

    if ppopt["CPF_PLOT_LEVEL"]:
        import matplotlib.pyplot as plt
        plt.show()
    sstats = dict(solver_name='PYPOWER', num_iters=cpf_results["iterations"])
    return results, success, sstats


def cpf_predictor(V, lam, Ybus, Sxfr, pv, pq,
                  step, z, Vprv, lamprv, parameterization):
    # sizes
    pvpq = np.r_[pv, pq]
    nb = len(V)
    npv = len(pv)
    npq = len(pq)

    # compute Jacobian for the power flow equations
    dSbus_dVm, dSbus_dVa = dSbus_dV(Ybus, V)

    j11 = dSbus_dVa[np.array([pvpq]).T, pvpq].real
    j12 = dSbus_dVm[np.array([pvpq]).T, pq].real
    j21 = dSbus_dVa[np.array([pq]).T, pvpq].imag
    j22 = dSbus_dVm[np.array([pq]).T, pq].imag

    J = sp.vstack([
        sp.hstack([j11, j12]),
        sp.hstack([j21, j22])
    ], format="csr")

    dF_dlam = -np.r_[Sxfr[pvpq].real, Sxfr[pq].imag].reshape((-1, 1))
    dP_dV, dP_dlam = cpf_p_jac(parameterization, z, V, lam, Vprv, lamprv, pv, pq)

    # linear operator for computing the tangent predictor
    J = sp.vstack([
        sp.hstack([J, dF_dlam]),
        sp.hstack([dP_dV, dP_dlam])
    ], format="csr")

    Vaprv = np.angle(V)
    Vmprv = abs(V)

    # compute normalized tangent predictor
    s = np.zeros(npv+2*npq+1)
    s[-1] = 1
    z[np.r_[pvpq, nb+pq, 2*nb]] = sp.linalg.spsolve(J, s)
    z = z / np.linalg.norm(z)

    Va0 = Vaprv
    Vm0 = Vmprv
    lam0 = lam

    # prediction for next step
    Va0[pvpq] = Vaprv[pvpq] + step * z[pvpq]
    Vm0[pq] = Vmprv[pq] + step * z[nb+pq]
    lam0 = lam + step * z[2*nb]
    V0 = Vm0 * np.exp((1j * Va0))

    return V0, lam0, z


def cpf_corrector(Ybus, Sbus, V0, ref, pv, pq,
                  lam0, Sxfr, Vprv, lamprv, z,
                  step, parameterization, ppopt):

    # default arguments
    if ppopt is None:
        ppopt = ppoption(ppopt)

    # options
    verbose = ppopt["VERBOSE"]
    tol = ppopt["PF_TOL"]
    max_it = ppopt["PF_MAX_IT"]

    # initialize
    converged = 0
    i = 0
    V = V0
    Va = np.angle(V)
    Vm = abs(V)
    lam = lam0

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)
    nb = len(V)
    j1 = 0
    j2 = npv    # j1:j2 - V angle of pv buses
    j3 = j2
    j4 = j2 + npq   # j1:j2 - V angle of pv buses
    j5 = j4
    j6 = j4 + npq   # j5:j6 - V mag of pq buses
    j7 = j6
    j8 = j6 + 1    # j7:j8 - lambda

    # evaluate F(x0, lam0), including Sxfr transfer/loading

    mis = V * np.conj(Ybus.dot(V)) - Sbus - lam*Sxfr
    F = np.r_[mis[pvpq].real, mis[pq].imag]

    # evaluate P(x0, lambda0)
    P = cpf_p(parameterization, step, z, V, lam, Vprv, lamprv, pv, pq)

    # augment F(x, lambda) with P(x, lambda)
    F = np.r_[F, P]

    # check tolerance
    normF = np.linalg.norm(F, inf)
    logger.debug('CPF correction')
    logger.debug('%2d: |F(x)| = %.10g', i, normF)
    if normF < tol:
        converged = 1

    # do Newton iterations
    while not converged and i < max_it:
        # update iteration counter
        i = i + 1

        # evaluate Jacobian
        dSbus_dVm, dSbus_dVa = dSbus_dV(Ybus, V)

        j11 = dSbus_dVa[np.array([pvpq]).T, pvpq].real
        j12 = dSbus_dVm[np.array([pvpq]).T, pq].real
        j21 = dSbus_dVa[np.array([pq]).T, pvpq].imag
        j22 = dSbus_dVm[np.array([pq]).T, pq].imag

        J = sp.vstack([
            sp.hstack([j11, j12]),
            sp.hstack([j21, j22])
        ], format="csr")

        dF_dlam = -np.r_[Sxfr[pvpq].real, Sxfr[pq].imag].reshape((-1, 1))
        dP_dV, dP_dlam = cpf_p_jac(parameterization, z, V, lam, Vprv, lamprv, pv, pq)

        # augment J with real/imag -Sxfr and z^T
        J = sp.vstack([
            sp.hstack([J, dF_dlam]),
            sp.hstack([dP_dV, dP_dlam])
        ], format="csr")

        # compute update step
        dx = -1 * sp.linalg.spsolve(J, F)

        # update voltage
        if npv:
            Va[pv] = Va[pv] + dx[j1:j2]
        if npq:
            Va[pq] = Va[pq] + dx[j3:j4]
            Vm[pq] = Vm[pq] + dx[j5:j6]
        V = Vm * np.exp((1j * Va))
        Vm = abs(V)
        Va = np.angle(V)

        # update lambda
        lam = lam + dx[j7:j8]

        # evalute F(x, lam)
        mis = V * np.conj(Ybus.dot(V)) - Sbus - lam*Sxfr
        F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]

        # evaluate P(x, lambda)
        P = cpf_p(parameterization, step, z, V, lam, Vprv, lamprv, pv, pq)

        # augment F(x, lambda) with P(x, lambda)
        F = np.r_[F, P]

        # check for convergence
        normF = np.linalg.norm(F, inf)
        logger.debug('%2d: |F(x)| = %.10g', i, normF)

        if normF < tol:
            converged = 1

    if not converged:
        logger.error('Newton''s method corrector did not converge in %d iterations.' % i)

    return V, converged, i, lam


def cpf_p_jac(parameterization, z, V, lam, Vprv, lamprv, pv, pq):
    if parameterization == 1:
        npv = len(pv)
        npq = len(pq)
        dP_dV = np.zeros(npv+2*npq)
        if lam >= lamprv:
            dP_dlam = 1.0
        else:
            dP_dlam = -1.0

    elif parameterization == 2:
        pvpq = np.r_[pv, pq]

        Va = np.angle(V)
        Vm = abs(V)
        Vaprv = np.angle(Vprv)
        Vmprv = abs(Vprv)
        dP_dV = 2 * (np.r_[Va[pvpq], Vm[pq]] -
                     np.r_[Vaprv[pvpq], Vmprv[pq]])
        if lam == lamprv:
            dP_dlam = 1.0
        else:
            dP_dlam = 2 * (lam - lamprv)

    elif parameterization == 3:
        nb = len(V)
        dP_dV = z[np.r_[pv, pq, nb+pq]]
        dP_dlam = z[2 * nb]

    return dP_dV, dP_dlam


def cpf_p(parameterization, step, z, V, lam, Vprv, lamprv, pv, pq):
    # evaluate P(x0, lambda0)
    if parameterization == 1:
        if lam >= lamprv:
            P = lam - lamprv - step
        else:
            P = lamprv - lam - step

    elif parameterization == 2:
        pvpq = np.r_[pv, pq]

        Va = np.angle(V)
        Vm = abs(V)
        Vaprv = np.angle(Vprv)
        Vmprv = abs(Vprv)
        P = sum(np.square(np.r_[Va[pvpq], Vm[pq], lam] -
                          np.r_[Vaprv[pvpq], Vmprv[pq], lamprv])) - np.square(step)

    elif parameterization == 3:
        pvpq = np.r_[pv, pq]

        nb = len(V)
        Va = np.angle(V)
        Vm = abs(V)
        Vaprv = np.angle(Vprv)
        Vmprv = abs(Vprv)
        P = np.dot(z[np.r_[pvpq, nb+pq, 2*nb]].reshape((1, -1)),
                   np.array(np.r_[Va[pvpq], Vm[pq], lam] - np.r_[Vaprv[pvpq], Vmprv[pq], lamprv]).T) - step

    return P



import warnings
import functools

from andes.utils.misc import elapsed


def timer(func):
    def wrapper(*args, **kwargs):
        t0, _ = elapsed()
        result = func(*args, **kwargs)
        _, s = elapsed(t0)
        return result, s
    return wrapper


def create_entry(*fields, three_params=True):
    """
    Helper function to create a list of fields for a model entry.

    Parameters
    ----------
    fields : tuple
        Additional fields to include in the list.
    three_params : bool, optional
        Whether to include 'idx', 'u', 'name' in the list (default is True).

    Returns
    -------
    list
        A list of fields for the model entry.
    """
    base_fields = ['idx', 'u', 'name'] if three_params else []
    return base_fields + list(fields)


def pretty_long_message(message, prefix="", max_length=80):
    """
    Pretty print a long message.

    Parameters
    ----------
    message : str
        The message to format.
    prefix : str, optional
        A prefix to add to each line of the message.
    max_length : int, optional
        The maximum length of each line.

    Returns
    -------
    str
        The formatted message.
    """
    if len(message) <= max_length:
        return message
    else:
        lines = [message[i:i+max_length] for i in range(0, len(message), max_length)]
        formatted_message = lines[0] + "\n" + "\n".join([prefix + line for line in lines[1:]])
        return formatted_message


def deprecated(message="This function is deprecated and will be removed in a future version."):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                message,
                category=DeprecationWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return decorator


deprec_get_idx = deprecated(
    "get_idx() is deprecated and will be removed in a future version. Please use get_all_idxes() instead.")



"""
PYPOWER utility functions.
"""
import logging  # NOQA
from copy import deepcopy  # NOQA

import numpy as np  # NOQA
from numpy import flatnonzero as find  # NOQA
import scipy.sparse as sp  # NOQA
from scipy.sparse import csr_matrix as c_sparse  # NOQA

from ams.pypower.idx import IDX  # NOQA

logger = logging.getLogger(__name__)


def bustypes(bus, gen):
    """
    Builds index lists of each type of bus (REF, PV, PQ).

    Generators with "out-of-service" status are treated as PQ buses with
    zero generation (regardless of Pg/Qg values in gen). Expects bus
    and gen have been converted to use internal consecutive bus numbering.

    Parameters
    ----------
    bus : ndarray
        Bus data.
    gen : ndarray
        Generator data.

    Returns
    -------
    ref : ndarray
        Index list of reference (REF) buses.
    pv : ndarray
        Index list of PV buses.
    pq : ndarray
        Index list of PQ buses.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    # get generator status
    nb = bus.shape[0]
    ng = gen.shape[0]
    # gen connection matrix, element i, j is 1 if, generator j at bus i is ON
    Cg = c_sparse((gen[:, IDX.gen.GEN_STATUS] > 0,
                   (gen[:, IDX.gen.GEN_BUS], range(ng))), (nb, ng))
    # number of generators at each bus that are ON
    bus_gen_status = (Cg * np.ones(ng, int)).astype(bool)

    # form index lists for slack, PV, and PQ buses
    ref = find((bus[:, IDX.bus.BUS_TYPE] == IDX.bus.REF) & bus_gen_status)  # ref bus index
    pv = find((bus[:, IDX.bus.BUS_TYPE] == IDX.bus.PV) & bus_gen_status)  # PV bus indices
    pq = find((bus[:, IDX.bus.BUS_TYPE] == IDX.bus.PQ) | ~bus_gen_status)  # PQ bus indices

    # pick a new reference bus if for some reason there is none (may have been
    # shut down)
    if (len(ref) == 0) & (len(pv) > 0):
        ref = np.zeros(1, dtype=int)
        ref[0] = pv[0]      # use the first PV bus
        pv = pv[1:]      # take it off PV list
    return ref, pv, pq


def sub2ind(shape, I, J, row_major=False):
    """
    Returns the linear indices of subscripts.

    Parameters
    ----------
    shape : tuple
        Shape of the grid or matrix.
    I : int
        Row subscript.
    J : int
        Column subscript.
    row_major : bool, optional
        If True, uses row-major order (default is False, using column-major order).

    Returns
    -------
    ind : int
        Linear index corresponding to the subscripts (I, J).
    """
    if row_major:
        ind = (I % shape[0]) * shape[1] + (J % shape[1])
    else:
        ind = (J % shape[1]) * shape[0] + (I % shape[0])

    return ind.astype(int)


def feval(func, *args, **kw_args):
    """
    Evaluates the function func using positional arguments args
    and keyword arguments kw_args.

    Parameters
    ----------
    func : str
        Name of the function to evaluate.
    *args : list
        Positional arguments for the function.
    **kw_args : dict
        Keyword arguments for the function.

    Returns
    -------
    result : any
        Result of evaluating the function.
    """
    return eval(func)(*args, **kw_args)


def have_fcn(name):
    """
    Checks if a Python module with the given name exists.

    Parameters
    ----------
    name : str
        Name of the Python module.

    Returns
    -------
    bool
        True if the module exists, False otherwise.
    """
    try:
        __import__(name)
        return True
    except ImportError:
        return False


def get_reorder(A, idx, dim=0):
    """
    Returns A with one of its dimensions indexed::

        B = get_reorder(A, idx, dim)

    Returns A[:, ..., :, idx, :, ..., :], where dim determines
    in which dimension to place the idx.

    @author: Ray Zimmerman (PSERC Cornell)
    """
    ndims = np.ndim(A)
    if ndims == 1:
        B = A[idx].copy()
    elif ndims == 2:
        if dim == 0:
            B = A[idx, :].copy()
        elif dim == 1:
            B = A[:, idx].copy()
        else:
            raise ValueError('dim (%d) may be 0 or 1' % dim)
    else:
        raise ValueError('number of dimensions (%d) may be 1 or 2' % dim)

    return B


def set_reorder(A, B, idx, dim=0):
    """
    Assigns B to A with one of the dimensions of A indexed.

    @return: A after doing A(:, ..., :, IDX, :, ..., :) = B
    where DIM determines in which dimension to place the IDX.

    @see: L{get_reorder}

    @author: Ray Zimmerman (PSERC Cornell)
    """
    A = A.copy()
    ndims = np.ndim(A)
    A = A.astype(B.dtype)
    if ndims == 1:
        A[idx] = B
    elif ndims == 2:
        if dim == 0:
            A[idx, :] = B
        elif dim == 1:
            A[:, idx] = B
        else:
            raise ValueError('dim (%d) may be 0 or 1' % dim)
    else:
        raise ValueError('number of dimensions (%d) may be 1 or 2' % dim)

    return A


def isload(gen):
    """
    Checks for dispatchable loads.

    Parameters
    ----------
    gen: np.ndarray
        The generator matrix.

    Returns
    -------
    array
        A column vector of 1's and 0's. The 1's correspond to rows of the
        C{gen} matrix which represent dispatchable loads. The current test is
        C{Pmin < 0 and Pmax == 0}. This may need to be revised to allow sensible
        specification of both elastic demand and pumped storage units.
    """
    return (gen[:, IDX.gen.PMIN] < 0) & (gen[:, IDX.gen.PMAX] == 0)


def hasPQcap(gen, hilo='B'):
    """
    Checks for P-Q capability curve constraints.

    Parameters
    ----------
    gen: np.ndarray
        The generator matrix.
    hilo : str, optional
        If 'U' this function returns C{True} only for rows corresponding to
        generators that require the upper constraint on Q.
        If 'L', only for those requiring the lower constraint.
        If not specified or has any other value it returns true for rows
        corresponding to gens that require either or both of the constraints.

    Returns
    -------
    array
        A column vector of 1's and 0's. The 1's correspond to rows of the
        C{gen} matrix which correspond to generators which have defined a
        capability curve (with sloped upper and/or lower bound on Q) and require
        that additional linear constraints be added to the OPF.

    Notes
    -----
        The C{gen} matrix in version 2 of the PYPOWER case format includes columns
        for specifying a P-Q capability curve for a generator defined as the
        intersection of two half-planes and the box constraints on P and Q.
        The two half planes are defined respectively as the area below the line
        connecting (Pc1, Qc1max) and (Pc2, Qc2max) and the area above the line
        connecting (Pc1, Qc1min) and (Pc2, Qc2min).

        It is smart enough to return C{True} only if the corresponding linear
        constraint is not redundant w.r.t the box constraints.
    """
    # check for errors capability curve data
    if np.any(gen[:, IDX.gen.PC1] > gen[:, IDX.gen.PC2]):
        logger.debug('hasPQcap: Pc1 > Pc2')
    if np.any(gen[:, IDX.gen.QC2MAX] > gen[:, IDX.gen.QC1MAX]):
        logger.debug('hasPQcap: Qc2max > Qc1max')
    if np.any(gen[:, IDX.gen.QC2MIN] < gen[:, IDX.gen.QC1MIN]):
        logger.debug('hasPQcap: Qc2min < Qc1min')

    L = np.zeros(gen.shape[0], bool)
    U = np.zeros(gen.shape[0], bool)
    k = np.nonzero(gen[:, IDX.gen.PC1] != gen[:, IDX.gen.PC2])

    if hilo != 'U':  # include lower constraint
        Qmin_at_Pmax = gen[k, IDX.gen.QC1MIN] + (gen[k, IDX.gen.PMAX] - gen[k, IDX.gen.PC1]) * (
            gen[k, IDX.gen.QC2MIN] - gen[k, IDX.gen.QC1MIN]) / (gen[k, IDX.gen.PC2] - gen[k, IDX.gen.PC1])
        L[k] = Qmin_at_Pmax > gen[k, IDX.gen.QMIN]

    if hilo != 'L':  # include upper constraint
        Qmax_at_Pmax = gen[k, IDX.gen.QC1MAX] + (gen[k, IDX.gen.PMAX] - gen[k, IDX.gen.PC1]) * (
            gen[k, IDX.gen.QC2MAX] - gen[k, IDX.gen.QC1MAX]) / (gen[k, IDX.gen.PC2] - gen[k, IDX.gen.PC1])
        U[k] = Qmax_at_Pmax < gen[k, IDX.gen.QMAX]

    return L | U


def fairmax(x):
    """
    Same as built-in C{max}, except breaks ties randomly.

    Takes a vector as an argument and returns the same output as the
    built-in function C{max} with two output parameters, except that
    where the maximum value occurs at more than one position in the
    vector, the index is chosen randomly from these positions as opposed
    to just choosing the first occurance.

    @see: C{max}

    @author: Ray Zimmerman (PSERC Cornell)
    """
    val = max(x)  # find max value
    i = np.nonzero(x == val)  # find all positions where this occurs
    n = len(i)  # number of occurences
    idx = i(np.fix(n * np.random()) + 1)  # select index randomly among occurances
    return val, idx



import unittest
import numpy as np

import ams


class TestDCOPF(unittest.TestCase):
    """
    Test routine `DCOPF`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.xlsx"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.DCOPF.init()
        self.assertTrue(self.ss.DCOPF.initialized, "DCOPF initialization failed!")

    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.DCOPF.update()
        self.ss.DCOPF.run(solver='CLARABEL')
        self.assertTrue(self.ss.DCOPF.converged, "DCOPF did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.DCOPF.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.DCOPF.update()
        self.ss.DCOPF.run(solver='CLARABEL')
        self.assertTrue(self.ss.DCOPF.converged, "DCOPF did not converge under line trip!")
        self.assertAlmostEqual(self.ss.DCOPF.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)  # reset

    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        # --- run DCOPF ---
        self.ss.DCOPF.run(solver='CLARABEL')
        pgs = self.ss.DCOPF.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.DCOPF.update()

        self.ss.DCOPF.run(solver='CLARABEL')
        pgs_pqt = self.ss.DCOPF.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.DCOPF.update()

        self.ss.DCOPF.run(solver='CLARABEL')
        pgs_pqt2 = self.ss.DCOPF.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")

    def test_dc2ac(self):
        """
        Test `DCOPF.dc2ac()` method.
        """
        self.ss.DCOPF.run(solver='CLARABEL')
        self.ss.DCOPF.dc2ac()
        self.assertTrue(self.ss.DCOPF.converted, "AC conversion failed!")
        self.assertTrue(self.ss.DCOPF.exec_time > 0, "Execution time is not greater than 0.")

        stg_idx = self.ss.StaticGen.get_all_idxes()
        pg_dcopf = self.ss.DCOPF.get(src='pg', attr='v', idx=stg_idx)
        pg_acopf = self.ss.ACOPF.get(src='pg', attr='v', idx=stg_idx)
        np.testing.assert_almost_equal(pg_dcopf, pg_acopf, decimal=3)

        bus_idx = self.ss.Bus.get_all_idxes()
        v_dcopf = self.ss.DCOPF.get(src='vBus', attr='v', idx=bus_idx)
        v_acopf = self.ss.ACOPF.get(src='vBus', attr='v', idx=bus_idx)
        np.testing.assert_almost_equal(v_dcopf, v_acopf, decimal=3)

        a_dcopf = self.ss.DCOPF.get(src='aBus', attr='v', idx=bus_idx)
        a_acopf = self.ss.ACOPF.get(src='aBus', attr='v', idx=bus_idx)
        np.testing.assert_almost_equal(a_dcopf, a_acopf, decimal=3)



import unittest
import numpy as np

import ams
from ams.shared import skip_unittest_without_MISOCP


class TestUC(unittest.TestCase):
    """
    Test routine `UC`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.xlsx"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])
        # run UC._initial_guess()
        self.off_gen = self.ss.UC._initial_guess()

    def test_initial_guess(self):
        """
        Test initial guess.
        """
        u_off_gen = self.ss.StaticGen.get(src='u', idx=self.off_gen)
        np.testing.assert_equal(u_off_gen, np.zeros_like(u_off_gen),
                                err_msg="UC._initial_guess() failed!")

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.UC.init()
        self.assertTrue(self.ss.UC.initialized, "UC initialization failed!")

    @skip_unittest_without_MISOCP
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        self.ss.UC.run()
        self.assertTrue(self.ss.UC.converged, "UC did not converge!")
        pg_off_gen = self.ss.UC.get(src='pg', attr='v', idx=self.off_gen)
        np.testing.assert_almost_equal(np.zeros_like(pg_off_gen),
                                       pg_off_gen, decimal=6,
                                       err_msg="Off generators are not turned off!")

    @skip_unittest_without_MISOCP
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        self.ss.UC.run()
        pgs = self.ss.UC.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.UC.update()

        self.ss.UC.run()
        pgs_pqt = self.ss.UC.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.alter(src='u', idx='PQ_2', value=0)
        self.ss.UC.update()

        self.ss.UC.run()
        pgs_pqt2 = self.ss.UC.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")

    @skip_unittest_without_MISOCP
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)
        self.ss.UC.update()

        self.ss.UC.run()
        self.assertTrue(self.ss.UC.converged, "UC did not converge under line trip!")
        plf_l3 = self.ss.UC.get(src='plf', attr='v', idx='Line_3')
        np.testing.assert_almost_equal(np.zeros_like(plf_l3),
                                       plf_l3, decimal=6)

        self.ss.Line.alter(src='u', idx='Line_3', value=1)


class TestUCDG(unittest.TestCase):
    """
    Test routine `UCDG`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.xlsx"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])
        # run `_initial_guess()`
        self.off_gen = self.ss.UCDG._initial_guess()

    def test_initial_guess(self):
        """
        Test initial guess.
        """
        u_off_gen = self.ss.StaticGen.get(src='u', idx=self.off_gen)
        np.testing.assert_equal(u_off_gen, np.zeros_like(u_off_gen),
                                err_msg="UCDG._initial_guess() failed!")

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.UCDG.init()
        self.assertTrue(self.ss.UCDG.initialized, "UCDG initialization failed!")

    @skip_unittest_without_MISOCP
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        self.ss.UCDG.run()
        self.assertTrue(self.ss.UCDG.converged, "UCDG did not converge!")
        pg_off_gen = self.ss.UCDG.get(src='pg', attr='v', idx=self.off_gen)
        np.testing.assert_almost_equal(np.zeros_like(pg_off_gen),
                                       pg_off_gen, decimal=6,
                                       err_msg="Off generators are not turned off!")

    @skip_unittest_without_MISOCP
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        self.ss.UCDG.run()
        pgs = self.ss.UCDG.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.UCDG.update()

        self.ss.UCDG.run()
        pgs_pqt = self.ss.UCDG.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.alter(src='u', idx='PQ_2', value=0)
        self.ss.UCDG.update()

        self.ss.UCDG.run()
        pgs_pqt2 = self.ss.UCDG.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")

    @skip_unittest_without_MISOCP
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)
        self.ss.UCDG.update()

        self.ss.UCDG.run()
        self.assertTrue(self.ss.UCDG.converged, "UCDG did not converge under line trip!")
        plf_l3 = self.ss.UCDG.get(src='plf', attr='v', idx='Line_3')
        np.testing.assert_almost_equal(np.zeros_like(plf_l3),
                                       plf_l3, decimal=6)

        self.ss.Line.alter(src='u', idx='Line_3', value=1)


class TestUCES(unittest.TestCase):
    """
    Test routine `UCES`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_uced_esd1.xlsx"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])
        # run `_initial_guess()`
        self.off_gen = self.ss.UCES._initial_guess()

    def test_initial_guess(self):
        """
        Test initial guess.
        """
        u_off_gen = self.ss.StaticGen.get(src='u', idx=self.off_gen)
        np.testing.assert_equal(u_off_gen, np.zeros_like(u_off_gen),
                                err_msg="UCES._initial_guess() failed!")

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.UCES.init()
        self.assertTrue(self.ss.UCES.initialized, "UCES initialization failed!")

    @skip_unittest_without_MISOCP
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        self.ss.UCES.run()
        self.assertTrue(self.ss.UCES.converged, "UCES did not converge!")
        pg_off_gen = self.ss.UCES.get(src='pg', attr='v', idx=self.off_gen)
        np.testing.assert_almost_equal(np.zeros_like(pg_off_gen),
                                       pg_off_gen, decimal=6,
                                       err_msg="Off generators are not turned off!")

    @skip_unittest_without_MISOCP
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        self.ss.UCES.run()
        pgs = self.ss.UCES.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.UCES.update()

        self.ss.UCES.run()
        pgs_pqt = self.ss.UCES.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.alter(src='u', idx='PQ_2', value=0)
        self.ss.UCES.update()

        self.ss.UCES.run()
        pgs_pqt2 = self.ss.UCES.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")

    @skip_unittest_without_MISOCP
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx=3, value=0)
        self.ss.UCES.update()

        self.ss.UCES.run()
        self.assertTrue(self.ss.UCES.converged, "UCES did not converge under line trip!")
        plf_l3 = self.ss.UCES.get(src='plf', attr='v', idx=3)
        np.testing.assert_almost_equal(np.zeros_like(plf_l3),
                                       plf_l3, decimal=6)

        self.ss.Line.alter(src='u', idx=3, value=1)
